try:
    import json
    import copy
    import h5py
    import random
    from pathlib import Path
    from .pointcloud_data import (
        concatenate_tensors,
        get_title_text,
        get_device,
        raise_maxpoint_warning,
        remap_labels,
        get_filter_index,
    )
    from torch.utils.data import DataLoader, Dataset, SequentialSampler
    from fastai.data_block import DataBunch
    from fastprogress.fastprogress import progress_bar
    import torch
    import numpy as np
    import types
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        from mmdet3d.core import LiDARInstance3DBoxes, Box3DMode
    from mmdet3d.core.points.lidar_points import LiDARPoints
    from mmdet3d.datasets.pipelines import Compose
    import plotly
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    from IPython.display import display
    from fastai.core import listify, recurse
    from fastai.torch_core import ifnone

    HAS_FASTAI = True
except Exception as e:
    HAS_FASTAI = False


class ODTransform3D(object):
    def __init__(
        self,
        rotation_range=[-0.78539816, 0.78539816],
        scaling_range=[0.95, 1.05],
        flip_x_prob=0.5,
        flip_y_prob=0.5,
    ):
        # In LIDAR coordinates, y (horizontal) and x (vertical) axis.
        self.tfms = [
            dict(
                type="RandomFlip3D",
                sync_2d=False,
                flip_ratio_bev_vertical=flip_x_prob,
                flip_ratio_bev_horizontal=flip_y_prob,
            ),
            dict(
                type="GlobalRotScaleTrans",
                rot_range=rotation_range,
                scale_ratio_range=scaling_range,
            ),
        ]

    def _create_transforms(self, point_cloud_range):
        self.tfms.append(
            dict(type="PointsRangeFilter", point_cloud_range=point_cloud_range)
        )
        self.tfms.append(
            dict(type="ObjectRangeFilter", point_cloud_range=point_cloud_range)
        )
        return Compose(self.tfms)


class PointCloudOD(Dataset):
    def __init__(
        self, path, folder="", class_mapping=None, transform_fn=False, **kwargs
    ):
        self.path = Path(path)
        with open(self.path / folder / "Statistics.json", "r") as f:
            self.statistics = json.load(f)
        self.features_to_keep = copy.copy(kwargs.get("extra_features", []))
        self.input_keys = kwargs.get("attributes", self.statistics["attributes"])
        self.input_keys.pop("rgbType", None)
        features_present = list(self.input_keys.keys())
        if "xyz" in features_present:
            features_present.remove("xyz")
        if self.features_to_keep != []:
            if not all([c in self.input_keys.keys() for c in self.features_to_keep]):
                raise Exception(
                    f"extra_features {self.features_to_keep} must be a subset of {features_present}"
                )
        self.features_to_keep += ["xyz"]
        self.input_keys = {
            k: v for k, v in self.input_keys.items() if k in self.features_to_keep
        }
        ## It is assumed that the pointcloud will have only X,Y & Z.
        extra_dim = (
            sum(
                [
                    len(v["max"]) if isinstance(v.get("max"), list) else 1
                    for _, v in self.input_keys.items()
                ]
            )
            - 3
        )
        self.extra_dim = extra_dim
        self.num_features = 3 + extra_dim  # XYZ + extra dimensions
        self.extra_features = self.input_keys

        self.scale_factor = self.statistics["parameters"]["scaleFactor"]
        self.block_size = self.statistics["parameters"]["tileSize"]

        self.max_point = self.statistics["parameters"]["maxPoints"]
        self.min_points = kwargs.get("min_points", None)
        if self.min_points is not None:
            assert (
                self.min_points < self.max_point
            ), f"min_points({self.min_points}) cannot be greater than max_points({self.max_point}) set during export"

        if "classification" in self.statistics["tileStatistics"]:
            # check if there is not any ground truth boxes in train/val folder
            if self.statistics["numberOfStoredOrientedBoundingBoxes"] == 0:
                raise Exception(
                    "Multipatch labels not found, data is not exported correctly."
                )

            class_info = self.statistics["tileStatistics"]["classification"]["table"]
            full_class_mapping = {
                int(c["classCode"]): str(c["classCode"]) for c in class_info
            }

            orig_classes = list(full_class_mapping.keys())
            self.remap_classes = kwargs.get("remap_classes", {})
            if not all([c in orig_classes for c in self.remap_classes.keys()]):
                raise Exception(
                    f"remap keys {list(self.remap_classes.keys())} must be a subset of {orig_classes}"
                )
            mapped_classes = set(self.remap_classes.values())
            unmapped_classes = set(
                [c for c in orig_classes if c not in self.remap_classes.keys()]
            )
            train_classes = sorted(mapped_classes.union(unmapped_classes))
            train_class_mapping = {int(c): str(c) for c in train_classes}

            self.classes_of_interest = kwargs.get("classes_of_interest", []).copy()

            if not all([c in train_classes for c in self.classes_of_interest]):
                raise Exception(
                    f"classes_of_interest {self.classes_of_interest} must be a subset of {train_classes}"
                )

            if class_mapping is None:
                class_mapping = train_class_mapping
            else:
                class_mapping = {
                    k: class_mapping.get(k, v) for k, v in train_class_mapping.items()
                }

            self.bg_code = kwargs.get("background_classcode", None)
            if self.classes_of_interest == [] and self.bg_code is not None:
                raise Exception(
                    "background_classcode can only be used when `classes_of_interest` is passed."
                )
            if self.bg_code is not None and type(self.bg_code) is not bool:
                raise Exception(
                    "Please enter a boolean value (True or False) for background_classcode."
                )

            if self.classes_of_interest != [] and self.bg_code:
                class_mapping = {
                    k: v
                    for k, v in class_mapping.items()
                    if k in self.classes_of_interest
                }

            class_mapping = {int(k): str(v) for k, v in class_mapping.items()}

            self.class_mapping = class_mapping

            self.c = len(self.class_mapping)
            self.classes = sorted(list(self.class_mapping.keys()))

            self.class2idx = {value: idx for idx, value in enumerate(self.classes)}
            self.idx2class = {idx: value for idx, value in enumerate(self.classes)}

            for k, v in self.remap_classes.items():
                if v in self.classes_of_interest:
                    self.classes_of_interest.append(k)
                    self.class2idx[k] = self.class2idx[v]
                if self.classes_of_interest == [] or not self.bg_code:
                    self.class2idx[k] = self.class2idx[v]

            self.classes_of_interest = sorted(list(set(self.classes_of_interest)))
            self.class2idx = {
                k: v for k, v in self.class2idx.items() if k in orig_classes
            }

            self.color_mapping = kwargs.get(
                "color_mapping",
                {
                    int(k): [random.choice(range(256)) for _ in range(3)]
                    for k, _ in self.class_mapping.items()
                },
            )

            # if classes is not continious classes = [1,3,8] or remapped
            self.remap = (
                self.classes != list(range(len(self.classes)))
                or self.remap_classes.__len__() != 0
            )

            remaped_train_classes = sorted(list(self.class2idx.keys()))

            self.z_range = self.statistics["tileStatistics"]["Z"]

            self.average_box_size = (
                np.array(
                    [
                        clas["orientedBoundingBoxAverageLength"]
                        for c in remaped_train_classes
                        for clas in class_info
                        if int(clas["classCode"]) == c
                    ]
                )
                * self.scale_factor
            ).tolist()
            # get the smallest box idx to calulate the voxel size
            box_idx = np.product(self.average_box_size, axis=1).argmin()
            box_size = self.average_box_size[box_idx]
            # taking 60 voxels in x and 20 voxels in z direction for each bbox
            self.voxel_size = [box_size[0] / 60, box_size[0] / 60, box_size[2] / 20]
            no_of_points = self.statistics["numberOfStoredRecords"]
            no_of_tiles = self.statistics["numberOfStoredTiles"]
            self.no_of_points_per_tile = no_of_points // no_of_tiles

            box_zminmax_range = [
                clas["orientedBoundingBoxZ"]
                for c in remaped_train_classes
                for clas in class_info
                if int(clas["classCode"]) == c
            ]

            box_zmin_range = [
                [box["min"][0], box["max"][0]] for box in box_zminmax_range
            ]

            anchor_zmin = [sum(box) / 2.0 for box in box_zmin_range]

            self.anchor_range = (
                np.array([[-1, -1, z_min, 1, 1, z_min] for z_min in anchor_zmin])
                * self.scale_factor
            ).tolist()

            if self.classes_of_interest != []:
                classes_of_interest = self.classes_of_interest
            else:
                classes_of_interest = None

            self._filter_box_point_percentage = kwargs.get(
                "filter_box_point_percentage", 0.2
            )
            self.transform = transform_fn
            if folder == "val":
                self._filter_box_point_percentage = 0.3
            if self.transform:
                if getattr(transform_fn, "_is_Transform3d", False):
                    self.transform = self.transform._detection_transforms()

                point_cloud_range = (
                    np.array([-1, -1, self.z_range["min"], 1, 1, self.z_range["max"]])
                    * self.scale_factor
                ).tolist()
                self.transform = self.transform._create_transforms(point_cloud_range)

        with h5py.File(self.path / folder / "ListTable.h5", "r") as f:
            self.tiles = f["Tiles"][:]
            relative_files = f["Files"][:]
            self.filenames = np.array([file.decode() for file in relative_files])

            if folder != "" and (
                self.min_points is not None or self.classes_of_interest != []
            ):
                if folder == "val":
                    self.min_points = None
                    classes_of_interest = None

                object_masks = f["ObjectMasks"][:]
                (
                    filterd_idx,
                    skip_block_min_points,
                    skip_block_COI,
                    _,
                ) = get_filter_index(
                    object_masks,
                    self.tiles,
                    self.min_points,
                    classes_of_interest,
                    True,
                )

                self._skip_block_min_points = skip_block_min_points
                self._skip_block_COI = skip_block_COI
                self._total_blocks = len(self.tiles)

                filterd_idx = np.sort(filterd_idx)
                self.tiles = self.tiles[filterd_idx]

                if len(self.tiles) == 0:
                    raise Exception(
                        f"The {folder} set is empty because everything "
                        "got filtered out."
                    )

        self.folder = folder
        if folder != "" and kwargs.get("filter_empty_tiles", False):
            self._filter()

    def _filter(self):
        idx = []
        for id, tile in enumerate(self.tiles):
            if tile[4] > 0:
                idx.append(id)
        self.tiles = self.tiles[idx]

    def __len__(self):
        return len(self.tiles)

    def __getitem__(self, tile_index):
        data = {}
        tile = self.tiles[tile_index]
        filename = self.path / self.folder / self.filenames[tile[0]]

        with h5py.File(filename, "r") as read_file:
            data["points"] = concatenate_tensors(
                read_file, self.input_keys, tile, pad=False
            )
            data["points"][:, :3] *= self.scale_factor

            if "orientedBoundingBox" in read_file.keys() and self.folder != "":
                data = self._get_bbox(data, read_file, tile)
                if self.transform and random.random() > 0.5:
                    data["points"] = LiDARPoints(
                        data["points"], points_dim=data["points"].shape[-1]
                    )
                    data["bbox3d_fields"] = ["gt_bboxes_3d"]
                    data["flip"] = False
                    data["flip_direction"] = None
                    data = self.transform(data)
                    data = self._post_process(data)
            else:
                data["tile_index"] = tile_index

        data["img_metas"] = dict(
            box_type_3d=LiDARInstance3DBoxes, box_mode_3d=Box3DMode.LIDAR
        )
        return data

    def _post_process(self, results):
        data = {}
        data["points"] = results["points"].tensor
        data["gt_bboxes_3d"] = results["gt_bboxes_3d"]
        data["gt_labels_3d"] = results["gt_labels_3d"]
        return data

    def _get_bbox(self, data, read_file, tile):
        bboxs = read_file["orientedBoundingBox"][tile[-2] : tile[-2] + tile[-1]].astype(
            np.float32
        )

        # filter boxes with less than 20% points in it of all the points in the box
        bbox_filter = self._get_valid_boxes(read_file, tile)
        bboxs = bboxs[bbox_filter]

        center = bboxs[:, [0, 1, 6]]
        length_width = 2 * bboxs[:, [4, 5]]
        height = (bboxs[:, 7] - bboxs[:, 6])[:, None]
        drx_dry = bboxs[:, [2, 3]]
        yaw = np.arctan2(drx_dry[:, 1], drx_dry[:, 0])[:, None]
        bboxs = np.concatenate(
            (
                center,
                length_width,
                height,
                yaw,
            ),
            axis=1,
        )

        bboxs[:, :6] *= self.scale_factor
        data["gt_bboxes_3d"] = LiDARInstance3DBoxes(bboxs, box_dim=7, with_yaw=True)

        bbox_labels = read_file.get("objectCode", None)
        if bbox_labels is not None:
            bbox_labels = torch.tensor(
                bbox_labels[tile[-2] : tile[-2] + tile[-1]]
            ).long()
            bbox_labels = bbox_labels[bbox_filter]
        else:
            bbox_labels = torch.tensor(np.zeros(center.shape[0])).long()

        # if train on only classes of interest
        if self.bg_code:
            filter_classes = self._get_class_of_interest_mask(bbox_labels)
            bbox_labels = bbox_labels[filter_classes]
            data["gt_bboxes_3d"] = data["gt_bboxes_3d"][filter_classes]

        if self.remap:
            bbox_labels = remap_labels(bbox_labels, self.class2idx).long()

        data["gt_labels_3d"] = bbox_labels
        return data

    def _get_class_of_interest_mask(self, bbox_labels):
        return torch.tensor([label in self.classes for label in bbox_labels]).bool()

    def _get_valid_boxes(self, read_file, tile):
        bbox_point_counts = read_file["orientedBoundingBoxCount"][
            tile[-2] : tile[-2] + tile[-1]
        ].astype(np.uint64)
        bbox_point_perecentage = bbox_point_counts[:, 0] / bbox_point_counts[:, 1]
        bbox_mask = bbox_point_perecentage > self._filter_box_point_percentage
        return bbox_mask

    def get_batch(self, batch_size, device=None):
        idxs = np.random.randint(0, len(self), batch_size)
        data_batch = []
        for idx in idxs:
            data_batch.append(self.__getitem__(idx))

        return to_device(collate_fn(data_batch), device)


def collate_fn(batch):
    return {k: [b[k] for b in batch] for k in batch[0].keys()}


defaults_device = (
    torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
)


def to_device(b, device=None):
    device = ifnone(device, defaults_device)
    return recurse(
        lambda x: x.to(device, non_blocking=True) if type(x) is torch.Tensor else x, b
    )


def proc_batch(self, x_batch):
    x_batch = to_device(x_batch, self.device)
    for f in listify(self.tfms):
        x_batch = f(x_batch)
    y_batch = dict(
        img_metas=np.array(x_batch["img_metas"]),
        boxes_3d=x_batch["gt_bboxes_3d"],
        labels_3d=x_batch["gt_labels_3d"],
    )
    return x_batch, y_batch


def show_batch(self, rows=2, color_mapping=None, **kwargs):
    """
    This can be used to visualize the exported dataset. Colors of the PointCloud
    are only used for better visualization, and it does not depict the
    actual classcode colors. Visualization of data, exported in a geographic
    coordinate system is not yet supported.
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional rows. Number of rows to show.
                            Default: ``2``.
    ---------------------   -------------------------------------------
    color_mapping           Optional dictionary. Mapping from object id
                            to RGB values. Colors of the PointCloud via
                            color_mapping are only used for better
                            visualization, and it does not depict the
                            actual classcode colors. Default value example:
                            {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    max_display_point       Optional integer. Maximum number of points
                            to display. Default is 20000. A warning will
                            be raised if the total points to display exceeds
                            this parameter. Setting this parameter will
                            randomly sample the specified number of points
                            and once set, it will be used for future uses.
    ---------------------   -------------------------------------------
    view_type               Optional string. Dataset type to display the
                            results.
                                * ``valid`` - For validation set.
                                * ``train`` - For training set.
                            Default: 'train'.
    =====================   ===========================================
    """
    max_display_point = kwargs.get("max_display_point", 20000)
    ds_type = kwargs.get("view_type", "train")
    color_mapping = self.color_mapping if color_mapping is None else color_mapping
    if ds_type == "train":
        ds = self.train_ds
    else:
        ds = self.valid_ds
    ids = list(range(len(ds)))
    random.shuffle(ids)
    for row_idx in range(rows):
        data_id = ids[row_idx]
        datapoint = ds[data_id]
        pc = datapoint["points"][:, :3]
        bbox = datapoint["gt_bboxes_3d"]
        bbox_class = datapoint["gt_labels_3d"]
        if pc.shape[0] > max_display_point:
            raise_maxpoint_warning(row_idx, kwargs, None, max_display_point)
            mask = torch.from_numpy(
                np.random.randint(0, pc.shape[0], max_display_point, dtype=np.int64)
            )
        else:
            mask = torch.arange(0, pc.shape[0])
        pc = pc[mask]
        x, y, z = pc.T

        scene = dict(aspectmode="data")
        layout = go.Layout(
            width=750,
            height=512,
            scene=scene,
        )

        color_list = np.tile(np.array([127, 127, 127]), (x.shape[0], 1)).astype(
            np.uint8
        )
        point_box_id = bbox.points_in_boxes_part(pc[:, :3].cuda()).cpu()

        mesh_objects = []
        for idx, coords in enumerate(bbox.corners):
            x_b, y_b, z_b = coords.T
            class_value = self.idx2class[bbox_class[idx].item()]
            class_name = self.class_mapping[class_value]
            class_name = f"class {class_name}"
            mesh_objects.append(
                go.Mesh3d(
                    x=x_b,
                    y=y_b,
                    z=z_b,
                    color="lightpink",
                    opacity=0.50,
                    alphahull=1,
                    text=class_name,
                )
            )

            color_list[point_box_id == idx] = color_mapping[class_value]

        fig = go.Figure(
            data=mesh_objects
            + [
                go.Scatter3d(
                    x=x,
                    y=y,
                    z=z,
                    mode="markers",
                    marker=dict(size=1, color=color_list),
                )
            ],
            layout=layout,
        )
        fig.show()


def pointcloud_od(
    path,
    class_mapping,
    batch_size,
    transform_fn,
    databunch_kwargs,
    **kwargs,
):
    del databunch_kwargs["bs"]
    env_device_type = str(databunch_kwargs.get("device"))
    if env_device_type == "cpu" or not torch.cuda.is_available():
        raise Exception(f"CPU is not supported for 'dataset_type':'PointCloudOD'.")

    train_dataset = PointCloudOD(
        path,
        folder="train",
        class_mapping=class_mapping,
        transform_fn=transform_fn,
        **kwargs,
    )
    val_dataset = PointCloudOD(
        path, folder="val", class_mapping=class_mapping, **kwargs
    )
    train_dl = DataLoader(
        train_dataset, collate_fn=collate_fn, batch_size=batch_size, **databunch_kwargs
    )
    valid_dl = DataLoader(
        val_dataset, collate_fn=collate_fn, batch_size=batch_size, **databunch_kwargs
    )
    device = get_device()
    data = DataBunch(train_dl, valid_dl, device=device, collate_fn=collate_fn)
    data.train_dl.proc_batch = types.MethodType(proc_batch, data.train_dl)
    data.valid_dl.proc_batch = types.MethodType(proc_batch, data.valid_dl)
    data.show_batch = types.MethodType(show_batch, data)
    block_z_range = [
        min(train_dataset.z_range["min"], val_dataset.z_range["min"]),
        max(train_dataset.z_range["max"], val_dataset.z_range["max"]),
    ]
    data.range = (
        np.array([-1, -1, block_z_range[0], 1, 1, block_z_range[1]]) * data.scale_factor
    ).tolist()

    data.path = data.train_ds.path
    data.chip_size = None
    data.chip_size = None
    data._image_space_used = None

    return data


def plot_results(
    row_idx,
    points,
    gt_boxes,
    gt_boxes_labels,
    gt_points_labels,
    pred_boxes,
    pred_boxes_labels,
    pred_points_labels,
    **kwargs,
):
    color_mapping = kwargs.get("color_mapping")
    class_mapping = kwargs.get("class_mapping")
    idx2class = kwargs.get("idx2class")
    save_html = kwargs.get("save_html", False)
    save_path = kwargs.get("save_path", ".")
    max_display_point = kwargs.get("max_display_point", 20000)
    fig = make_subplots(rows=1, cols=2, specs=[[{"type": "scene"}, {"type": "scene"}]])
    boxes_3d = [gt_boxes, pred_boxes]
    labels_3d = [gt_boxes_labels, pred_boxes_labels]
    points_labels_3d = [gt_points_labels, pred_points_labels]
    x, y, z = points.T
    for col_idx, (boxes, box_labels, point_labels) in enumerate(
        zip(boxes_3d, labels_3d, points_labels_3d)
    ):
        color_list = np.tile(np.array([127, 127, 127]), (points.shape[0], 1)).astype(
            np.uint8
        )
        for box_idx, coords in enumerate(boxes.corners):
            x_b, y_b, z_b = coords.detach().cpu().T
            class_value = idx2class[box_labels[box_idx].item()]
            class_name = class_mapping[class_value]
            if col_idx == 1:
                class_name = f"pred_class {class_name}"
            else:
                class_name = f"class {class_name}"
            fig.add_trace(
                go.Mesh3d(
                    x=x_b,
                    y=y_b,
                    z=z_b,
                    color="lightpink",
                    opacity=0.50,
                    alphahull=1,
                    text=class_name,
                ),
                row=1,
                col=col_idx + 1,
            )
            color_list[point_labels == box_idx] = color_mapping[class_value]
        pc_object = go.Scatter3d(
            x=x,
            y=y,
            z=z,
            mode="markers",
            marker=dict(size=1, color=color_list),
        )
        fig.add_trace(pc_object, row=1, col=col_idx + 1)

    title_text = get_title_text(row_idx, save_html, max_display_point)
    fig.update_layout(
        scene=dict(aspectmode="data"),
        scene2=dict(aspectmode="data"),
        title_text=title_text,
        width=kwargs.get("width", 750),
        height=kwargs.get("height", 512),
        showlegend=False,
        title_x=0.5,
    )

    fig2 = go.FigureWidget(fig)

    def cam_change(layout, camera):
        if fig2.layout.scene2.camera == camera:
            return
        else:
            fig2.layout.scene2.camera = camera

    def cam_change2(layout, camera):
        if fig2.layout.scene.camera == camera:
            return
        else:
            fig2.layout.scene.camera = camera

    fig2.layout.scene.on_change(cam_change, "camera")
    fig2.layout.scene2.on_change(cam_change2, "camera")
    if save_html:
        save_path = Path(save_path)
        plotly.io.write_html(fig, str(save_path / "show_results.html"))
        return
    else:
        display(fig2)


def confusion_matrix3d(pred, target, n_gts, classes, iou_thresh=0.1):
    tps, p_clas, p_scores = [], [], []
    for idx in range(len(pred)):
        pred_bboxes, pred_labels, pred_scores = (
            pred[idx]["boxes_3d"],
            pred[idx]["labels_3d"].detach().cpu(),
            pred[idx]["scores_3d"].detach().cpu(),
        )
        tgt_bboxes, tgt_lables = (
            target["boxes_3d"][idx],
            target["labels_3d"][idx].detach().cpu(),
        )

        if len(pred_labels) != 0 and len(tgt_lables) != 0:
            ious = pred_bboxes.overlaps(pred_bboxes, tgt_bboxes)
            max_iou, matches = ious.max(1)
            detected = []
            for i in range(len(pred_labels)):
                if (
                    max_iou[i] >= iou_thresh
                    and matches[i] not in detected
                    and tgt_lables[matches[i]] == pred_labels[i]
                ):
                    detected.append(matches[i])
                    tps.append(1)
                else:
                    tps.append(0)
            p_clas.append(pred_labels)
            p_scores.append(pred_scores)
            n_gts += ((tgt_lables[:, None]) == classes[None, :]).sum(0)

    return tps, p_scores, p_clas, n_gts


def predict_h5(self, path, output_path, **kwargs):
    self._free_memory()
    path = Path(path)
    if output_path is None:
        output_path = path.parent / "results"
    else:
        output_path = Path(output_path)
    progressor = kwargs.pop("progressor", None)
    batch_size = kwargs.get("batch_size", 1)

    extra_features = self._data.features_to_keep.copy()
    if "xyz" in extra_features:
        extra_features.remove("xyz")

    data = PointCloudOD(path, extra_features=extra_features)

    if progressor is not None:
        progressor.set_total_blocks(len(data))

    output_path.mkdir(parents=True, exist_ok=True)
    sampler = SequentialSampler(data)

    dataloader = DataLoader(
        data,
        collate_fn=collate_fn,
        batch_size=batch_size,
        sampler=sampler,
    )
    predict_batch_h5(self, dataloader, output_path, progressor, **kwargs)
    self._reset_thresh()

    return output_path


def predict_batch_h5(self, dl, output_path, progressor, **kwargs):
    detect_thresh = kwargs.get("detect_thresh", 0.1)
    nms_overlap = kwargs.get("nms_overlap", 0.6)
    current_file_name = ""
    for data in progress_bar(dl):
        tile_index = data.pop("tile_index")
        data = to_device(data)
        pred = self._pred_batch(data, detect_thresh, nms_overlap)

        tile = dl.dataset.tiles[tile_index]
        if len(tile.shape) < 2:
            tile = tile[None]
        fname = np.array(dl.dataset.filenames)[tile[:, 0]]
        _, unique_index = np.unique(fname, return_index=True)
        # get unique name form sorted index in actual array
        fname = [fname[i] for i in np.sort(unique_index)]
        # add batch_size for spliting prediction till last batch number
        unique_index = list(np.sort(unique_index)) + [dl.batch_size]
        for i, ufname in enumerate(fname):
            ufname = dl.dataset.path / dl.dataset.folder / ufname
            if ufname != current_file_name:
                current_file_name = ufname
                h5_file = h5py.File(current_file_name, "r")
                no_of_point, _ = h5_file["xyz"].shape
                h5_file.close()
                boxes_pred = []
                labels_pred = []
                confidence_pred = []
                boxes_tile_no = []
                low = high = 0

            predictions = split_prediction(
                self,
                pred[unique_index[i] : unique_index[i + 1]],
                data["points"][unique_index[i] : unique_index[i + 1]],
                tile_index[unique_index[i] : unique_index[i + 1]],
            )

            boxes_pred.extend(predictions[0])
            labels_pred.extend(predictions[1])
            confidence_pred.extend(predictions[2])
            boxes_tile_no.extend(predictions[3])
            high = low + predictions[4]

            if high == no_of_point:
                save_h5(
                    output_path / dl.dataset.filenames[int(tile[unique_index[i]][0])],
                    np.array(boxes_pred),
                    np.array(labels_pred, dtype=np.uint8),
                    np.array(confidence_pred),
                    np.array(boxes_tile_no, dtype=np.uint64),
                )
            low = high

        if progressor is not None:
            progressor.current_block(tile_index[0])


def export_boxes(bboxes, scale_factor):
    """
    input: [center_x, center_y, z_min, x_size, y_size, z_size, yaw]
    return: [center_x, center_y, direction_x, direction_y, half_size_x, half_size_y, z_min, z_max]
    """
    bboxes[:, :6] /= scale_factor
    cenetr_xy = bboxes[:, [0, 1]]
    z_min = bboxes[:, [2]]
    z_max = z_min + bboxes[:, [5]]
    half_size_xy = bboxes[:, [3, 4]] / 2
    direction_x, direction_y = np.cos(bboxes[:, [6]]), np.sin(bboxes[:, [6]])

    return np.concatenate(
        (cenetr_xy, direction_x, direction_y, half_size_xy, z_min, z_max), axis=1
    ).tolist()


def split_prediction(model, preds, points, tile_index):
    batch_export_bboxs = []
    batch_labels = []
    batch_confidance = []
    batch_box_tiles = []
    no_of_points_in_batch = 0
    for idx, pred in enumerate(preds):
        boxes = pred["boxes_3d"]
        labels = pred["labels_3d"]
        scores = pred["scores_3d"]
        no_of_points_in_batch += points[idx].shape[0]

        batch_export_bboxs.extend(export_boxes(boxes.tensor, model._data.scale_factor))
        batch_labels.extend(
            np.array(list(model._data.idx2class.values()))[labels.tolist()]
        )
        batch_confidance.extend(scores.tolist())
        batch_box_tiles.extend([tile_index[idx]] * scores.shape[0])

    return (
        batch_export_bboxs,
        batch_labels,
        batch_confidance,
        batch_box_tiles,
        no_of_points_in_batch,
    )


def save_h5(filename, boxes_pred, labels_pred, confidences_pred, boxes_tile_no):
    filename = Path(filename)
    if not filename.parent.exists():
        filename.parent.mkdir(parents=True, exist_ok=True)

    filename_pred = filename.parent / (filename.stem + "_pred.h5")
    with h5py.File(filename_pred, "w") as file:
        file.create_dataset("pred_boxes", data=boxes_pred)
        file.create_dataset("pred_labels", data=labels_pred)
        file.create_dataset("pred_confidence", data=confidences_pred)
        file.create_dataset("pred_boxes_block", data=boxes_tile_no)



import os
import tempfile
import random
import traceback
import json
import warnings
import math
from pathlib import Path

import_exception = None

import arcgis
from arcgis.features import FeatureLayer

try:
    from ._arcgis_model import ArcGISModel, _raise_fastai_import_error
    from fastai.tabular import (
        tabular_learner,
        TabularDataBunch,
        TabularList,
        TabularModel,
    )
    from fastai.tabular.transform import FillMissing, Categorify, Normalize
    from fastai.basic_train import Learner, load_learner
    from fastprogress.fastprogress import progress_bar
    from .._utils.tabular_data import TabularDataObject, explain_prediction, add_h3
    from .._utils.common import _get_emd_path
    from fastai.torch_core import split_model_idx
    import torch
    from fastai.metrics import r2_score

    HAS_FASTAI = True

except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

try:
    import numpy as np

    HAS_NUMPY = True
except:
    HAS_NUMPY = False


def _get_learner_object(data, layers, emb_szs, ps, emb_drop, pretrained_path):
    if pretrained_path:
        learn = load_learner(
            os.path.dirname(pretrained_path),
            os.path.basename(pretrained_path).rsplit(".", 1)[0] + "_exported.pth",
        )
        learn.path = data.path
        if not data._is_empty:
            learn.data = data._databunch
    else:
        databunch = data._databunch
        if not emb_szs or isinstance(emb_szs, dict):
            emb_szs = databunch.get_emb_szs({} if not emb_szs else emb_szs)

        model = TabularModel(
            emb_szs,
            len(databunch.cont_names),
            out_sz=databunch.c,
            layers=layers,
            ps=ps,
            emb_drop=emb_drop,
            y_range=None,
            use_bn=False,
        )
        learn = Learner(databunch, model, path=data.path)

    return learn


class FullyConnectedNetwork(ArcGISModel):
    """
    Creates a :class:`~arcgis.learn.FullyConnectedNetwork` Object.
    Based on the Fast.ai's Tabular Learner

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required TabularDataObject. Returned data object from
                            :class:`~arcgis.learn.prepare_tabulardata` function.
    ---------------------   -------------------------------------------
    layers                  Optional list, specifying the number of nodes in each layer.
                            Default: [500, 100] is used.
                            2 layers each with nodes 500 and 100 respectively.
    ---------------------   -------------------------------------------
    emb_szs                 Optional dict, variable name with embedding size
                            for categorical variables.
                            If not specified, then calculated using fastai.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.FullyConnectedNetwork` Object
    """

    def __init__(self, data, layers=None, emb_szs=None, **kwargs):
        if data._is_unsupervised:
            raise Exception("Cannot train on unsupervised data")

        super().__init__(data, None)

        if isinstance(self._data._dependent_variable, list):
            self._data._dependent_variable = self._data._dependent_variable[0]

        self._backbone = None
        if layers is None:
            layers = [500, 100]

        ps = kwargs.get("ps", None)
        emb_drop = kwargs.get("emb_drop", 0)

        self.learn = _get_learner_object(
            data, layers, emb_szs, ps, emb_drop, kwargs.get("pretrained_path", None)
        )
        self._layers = layers
        self.learn.model = self.learn.model.to(self._device)
        idx = 1
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=1e-03)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.FullyConnectedNetwork` Object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :class:`~arcgis.learn.prepare_tabulardata` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.FullyConnectedNetwork` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        dependent_variable = emd["dependent_variable"]
        categorical_variables = emd["categorical_variables"]
        continuous_variables = emd["continuous_variables"]
        _is_classification = False
        if emd["_is_classification"] == "classification":
            _is_classification = True

        layers = emd["layers"]
        cell_sizes = emd.get("cell_sizes", None)
        if data is None:
            data = TabularDataObject._empty(
                categorical_variables, continuous_variables, dependent_variable, None
            )
            data._is_classification = _is_classification
            data._cell_sizes = cell_sizes
            class_object = cls(data, pretrained_path=emd_path)
            class_object._data.emd = emd
            class_object._data.emd_path = emd_path
            return class_object

        return cls(data, layers=layers, pretrained_path=str(emd_path))

    def save(
        self,
        name_or_path,
        framework="PyTorch",
        publish=False,
        gis=None,
        save_optimizer=False,
        **kwargs,
    ):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by :class:`~arcgis.learn.SingleShotDetector`, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` can be
                                passed as an optional keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
        =====================   ===========================================
        """

        if "\\" in name_or_path or "/" in name_or_path:
            path = os.path.abspath(name_or_path)
        else:
            path = os.path.join(self._data.path, "models", name_or_path)
            if not os.path.exists(os.path.dirname(path)):
                os.mkdir(os.path.dirname(path))

        if not os.path.exists(path):
            os.mkdir(path)

        self.learn.export(os.path.join(path, os.path.basename(path) + "_exported.pth"))
        # from IPython.utils import io

        # with io.capture_output() as captured:
        super().save(
            path, framework, publish, gis, save_optimizer=save_optimizer, **kwargs
        )
        # print(captured.stdout)
        return Path(path)

    @property
    def _model_metrics(self):
        from IPython.utils import io

        with io.capture_output() as captured:
            score = self.score()

        return {"score": score}

    @property
    def feature_importances_(self):
        """
        :return: the global feature importance summary plot from SHAP.Feature is temporarily disabled.
        """
        warnings.warn(
            "Feature importance for Fully Connected Network is currently disabled due to package incompatibility and is under review"
        )
        # processed_dataframe = None
        # explain_index = None
        # random_index = None
        # explain_prediction(
        #    self,
        #    processed_dataframe,
        #    index=explain_index,
        #    random_index=random_index,
        #    predictor=None,
        #    global_pred=True,
        # )
        return

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["ModelType"] = "FullyConnectedNetwork"
        _emd_template["layers"] = self._layers
        _emd_template["dependent_variable"] = self._data._dependent_variable
        _emd_template["categorical_variables"] = self._data._categorical_variables
        _emd_template["continuous_variables"] = self._data._continuous_variables
        _emd_template["_is_classification"] = (
            "classification" if self._data._is_classification else "regression"
        )
        _emd_template["cell_sizes"] = self._data._cell_sizes

        return _emd_template

    def _predict(self, data_frame_row):
        return self.learn.predict(data_frame_row)

    def _df_predict(self, dataframe):
        # from fastai.data_block import split_kwargs_by_func, grab_idx, DatasetType
        # ds_type = DatasetType.Valid
        # ds = self.learn.dl(ds_type).dataset
        #
        # current_databunch = self.learn.data
        # databunch_half, databunch_second_half = self._data._prepare_validation_databunch(dataframe)
        # return dataframe, databunch_half, databunch_second_half
        # try:
        #     self.learn.data = databunch
        #     preds = self.learn.get_preds(ds_type)[0]
        # except Exception as e:
        #     raise e
        # finally:
        #     self.learn.data = current_databunch
        #
        # analyze_kwargs, kwargs = split_kwargs_by_func({}, ds.y.analyze_pred)
        # preds = [ds.y.analyze_pred(grab_idx(preds, i), **analyze_kwargs) for i in range(len(preds))]
        # preds = [ds.y.reconstruct(z) for z in preds]
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        preds = []

        for i in progress_bar(range(len(dataframe))):
            prediction = self._predict(dataframe.iloc[i])[0].obj
            if isinstance(prediction, (list, np.ndarray)):
                prediction = prediction[0]
            if isinstance(prediction, np.float32):
                prediction = prediction.astype(np.float64)
            preds.append(prediction)

        return preds

    def predict(
        self,
        input_features=None,
        explanatory_rasters=None,
        datefield=None,
        distance_features=None,
        output_layer_name="Prediction Layer",
        gis=None,
        prediction_type="features",
        output_raster_path=None,
        match_field_names=None,
        explain=False,
        explain_index=None,
    ):
        """

        Predict on data from feature layer, dataframe and or raster data.

        =================================   =========================================================================
        **Parameter**                        **Description**
        ---------------------------------   -------------------------------------------------------------------------
        input_features                      Optional :class:`~arcgis.features.FeatureLayer` or spatially enabled dataframe.
                                            Required if prediction_type='features'.
                                            Contains features with location and
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        explanatory_rasters                 Optional list of Raster Objects.
                                            If prediction_type='raster', must contain all rasters
                                            required to make predictions.
        ---------------------------------   -------------------------------------------------------------------------
        datefield                           Optional string. Field name from feature layer
                                            that contains the date, time for the input features.
                                            Same as :meth:`~arcgis.learn.prepare_tabulardata` .
        ---------------------------------   -------------------------------------------------------------------------
        distance_features                   Optional List of :class:`~arcgis.features.FeatureLayer` objects.
                                            These layers are used for calculation of field "NEAR_DIST_1",
                                            "NEAR_DIST_2" etc in the output dataframe.
                                            These fields contain the nearest feature distance
                                            from the input_features.
                                            Same as :meth:`~arcgis.learn.prepare_tabulardata` .
        ---------------------------------   -------------------------------------------------------------------------
        output_layer_name                   Optional string. Used for publishing the output layer.
        ---------------------------------   -------------------------------------------------------------------------
        gis                                 Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                            If not specified then active gis user is taken.
        ---------------------------------   -------------------------------------------------------------------------
        prediction_type                     Optional String.
                                            Set 'features' or 'dataframe' to make output feature layer predictions.
                                            With this feature_layer argument is required.

                                            Set 'raster', to make prediction raster.
                                            With this rasters must be specified.
        ---------------------------------   -------------------------------------------------------------------------
        output_raster_path                  Optional path.
                                            Required when prediction_type='raster', saves
                                            the output raster to this path.
        ---------------------------------   -------------------------------------------------------------------------
        match_field_names                   Optional dictionary.
                                            Specify mapping of field names from prediction set
                                            to training set.
                                            For example:

                                                | {
                                                |    "Field_Name_1": "Field_1",
                                                |    "Field_Name_2": "Field_2"
                                                | }
        ---------------------------------   -------------------------------------------------------------------------
        explain                             Optional Bool.
                                            Setting this parameter to true generates prediction explaination plot.
                                            Plot is generated using model interpretability library called SHAP.
                                            (https://github.com/slundberg/shap). Feature is temporarily disabled.
        ---------------------------------   -------------------------------------------------------------------------
        explain_index                       Optional Int.
                                            The index of the dataframe passed to the predict function for which model
                                            interpretability is desired. If the parameter is not passed and if the
                                            explain parameter is set to true, the SHAP plot will be generated for a
                                            random index of the dataframe.
        =================================   =========================================================================

        :return: Feature Layer if prediction_type='features', dataframe for prediction_type='dataframe' else creates an output raster.

        """

        rasters = explanatory_rasters if explanatory_rasters else []
        if explain:
            # try:
            #     import shap
            # except:
            #     warnings.warn(
            #         "Prediction cannot be explained as SHAP is not installed. Please install SHAP to get explainability working."
            #     )
            #     explain = False
            #     explain_index = None
            warnings.warn(
                "Model explainability feature for Fully Connected Network is currently disabled due to package incompatibility and is under review"
            )
            explain = False
            explain_index = None
        if prediction_type in ["features", "dataframe"]:
            if input_features is None:
                raise Exception("Feature Layer required for predict_features=True")

            gis = gis if gis else arcgis.env.active_gis
            return self._predict_features(
                input_features,
                rasters,
                datefield,
                distance_features,
                output_layer_name,
                gis,
                match_field_names,
                prediction_type,
                explain,
                explain_index,
            )
        else:
            if not rasters:
                raise Exception("Rasters required for predict_features=False")

            if not output_raster_path:
                raise Exception(
                    "Please specify output_raster_folder_path to save the output."
                )

            return self._predict_rasters(
                output_raster_path, rasters, match_field_names, explain, explain_index
            )

    def _predict_features(
        self,
        input_features,
        rasters=None,
        datefield=None,
        distance_feature_layers=None,
        output_name="Prediction Layer",
        gis=None,
        match_field_names=None,
        prediction_type="features",
        explain=False,
        explain_index=None,
    ):
        cell_sizes = self._data._cell_sizes
        if isinstance(input_features, FeatureLayer):
            if cell_sizes and not rasters:
                dataframe = input_features.query(out_sr=4326).sdf
                dataframe = add_h3(dataframe, cell_sizes)
            else:
                dataframe = input_features.query().sdf
        else:
            dataframe = input_features.copy()

        fields_needed = (
            self._data._categorical_variables + self._data._continuous_variables
        )
        distance_feature_layers = (
            distance_feature_layers if distance_feature_layers else []
        )

        continuous_variables = self._data._continuous_variables

        feature_layer_columns = []
        for column in dataframe.columns:
            column_name = column
            categorical = False

            if column_name in fields_needed:
                if column_name not in continuous_variables:
                    categorical = True
            elif match_field_names and match_field_names.get(column_name):
                if match_field_names.get(column_name) not in continuous_variables:
                    categorical = True
            else:
                continue

            feature_layer_columns.append((column_name, categorical))

        raster_columns = []
        if rasters:
            for raster in rasters:
                column_name = raster.name
                categorical = False
                if column_name in fields_needed:
                    if column_name not in continuous_variables:
                        categorical = True
                elif match_field_names and match_field_names.get(column_name):
                    column_name = match_field_names.get(column_name)
                    if column_name not in continuous_variables:
                        categorical = True
                else:
                    continue

                raster_columns.append((raster, categorical))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            (
                processed_dataframe,
                fields_mapping,
            ) = TabularDataObject._prepare_dataframe_from_features(
                input_features,
                self._data._dependent_variable,
                feature_layer_columns,
                raster_columns,
                datefield,
                cell_sizes,
                distance_feature_layers,
            )

        if match_field_names:
            processed_dataframe.rename(columns=match_field_names, inplace=True)

        for field in fields_needed:
            if field not in processed_dataframe.columns:
                raise Exception(f"Field missing {field}")
        if explain:
            if explain_index is None:
                random_index = True
            else:
                random_index = False
            explain_prediction(
                self,
                processed_dataframe,
                index=explain_index,
                random_index=random_index,
                predictor=None,
            )

        dataframe["prediction_results"] = self._df_predict(processed_dataframe.copy())
        if prediction_type == "dataframe":
            return dataframe

        if "SHAPE" in list(dataframe.columns):
            return dataframe.spatial.to_featurelayer(output_name, gis)
        else:
            import tempfile

            with tempfile.TemporaryDirectory() as tmpdir:
                table_file = os.path.join(tmpdir, output_name + ".xlsx")
                dataframe.to_excel(table_file, index=False, header=True)
                online_table = gis.content.add(
                    {"type": "Microsoft Excel", "overwrite": True}, table_file
                )
                return online_table.publish(overwrite=True)

    def _predict_rasters(
        self,
        output_folder_path,
        rasters,
        match_field_names=None,
        explain=False,
        explain_index=None,
    ):
        if not os.path.exists(os.path.dirname(output_folder_path)):
            raise Exception("Output directory doesn't exist")

        if os.path.exists(output_folder_path):
            raise Exception("Output Folder already exists")

        try:
            import arcpy
        except:
            raise Exception("This function requires arcpy.")

        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")

        fields_needed = (
            self._data._continuous_variables + self._data._categorical_variables
        )

        try:
            arcpy.env.outputCoordinateSystem = rasters[0].extent["spatialReference"][
                "wkt"
            ]
        except:
            arcpy.env.outputCoordinateSystem = rasters[0].extent["spatialReference"][
                "wkid"
            ]

        xmin = rasters[0].extent["xmin"]
        xmax = rasters[0].extent["xmax"]
        ymin = rasters[0].extent["ymin"]
        ymax = rasters[0].extent["ymax"]
        min_cell_size_x = rasters[0].mean_cell_width
        min_cell_size_y = rasters[0].mean_cell_height

        default_sr = rasters[0].extent["spatialReference"]

        for raster in rasters:
            point_upper = arcgis.geometry.Point(
                {
                    "x": raster.extent["xmin"],
                    "y": raster.extent["ymax"],
                    "sr": raster.extent["spatialReference"],
                }
            )
            point_lower = arcgis.geometry.Point(
                {
                    "x": raster.extent["xmax"],
                    "y": raster.extent["ymin"],
                    "sr": raster.extent["spatialReference"],
                }
            )
            cell_size = arcgis.geometry.Point(
                {
                    "x": raster.mean_cell_width,
                    "y": raster.mean_cell_height,
                    "sr": raster.extent["spatialReference"],
                }
            )

            points = arcgis.geometry.project(
                [point_upper, point_lower, cell_size],
                raster.extent["spatialReference"],
                default_sr,
            )
            point_upper = points[0]
            point_lower = points[1]
            cell_size = points[2]

            if xmin > point_upper.x:
                xmin = point_upper.x
            if ymax < point_upper.y:
                ymax = point_upper.y
            if xmax < point_lower.x:
                xmax = point_lower.x
            if ymin > point_lower.y:
                ymin = point_lower.y

            if min_cell_size_x > cell_size.x:
                min_cell_size_x = cell_size.x

            if min_cell_size_y > cell_size.y:
                min_cell_size_y = cell_size.y

        max_raster_columns = int(abs(math.ceil((xmax - xmin) / min_cell_size_x)))
        max_raster_rows = int(abs(math.ceil((ymax - ymin) / min_cell_size_y)))

        point_upper = arcgis.geometry.Point({"x": xmin, "y": ymax, "sr": default_sr})
        cell_size = arcgis.geometry.Point(
            {"x": min_cell_size_x, "y": min_cell_size_y, "sr": default_sr}
        )

        raster_data = {}
        for raster in rasters:
            field_name = raster.name
            point_upper_translated = arcgis.geometry.project(
                [point_upper], default_sr, raster.extent["spatialReference"]
            )[0]
            cell_size_translated = arcgis.geometry.project(
                [cell_size], default_sr, raster.extent["spatialReference"]
            )[0]
            if field_name in fields_needed:
                raster_read = raster.read(
                    origin_coordinate=(
                        point_upper_translated.x,
                        point_upper_translated.y,
                    ),
                    ncols=max_raster_columns,
                    nrows=max_raster_rows,
                    cell_size=(cell_size_translated.x, cell_size_translated.y),
                )
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f"_{index}"
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            elif match_field_names and match_field_names.get(raster.name):
                field_name = match_field_names.get(raster.name)
                raster_read = raster.read(
                    origin_coordinate=(
                        point_upper_translated.x,
                        point_upper_translated.y,
                    ),
                    ncols=max_raster_columns,
                    nrows=max_raster_rows,
                    cell_size=(cell_size_translated.x, cell_size_translated.y),
                )
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f"_{index}"
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            else:
                continue

        for field in fields_needed:
            if (
                field not in list(raster_data.keys())
                and match_field_names
                and match_field_names.get(field, None) is None
            ):
                raise Exception(f"Field missing {field}")

        processed_data = []

        length_values = len(raster_data[list(raster_data.keys())[0]])
        for i in range(length_values):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_data.append(processed_row)
        if explain:
            if explain_index is None:
                random_index = True
            else:
                random_index = False
            explain_prediction(
                self,
                pd.DataFrame(
                    data=np.array(processed_data), columns=sorted(raster_data)
                ),
                index=explain_index,
                random_index=random_index,
                predictor=None,
            )

        processed_numpy = np.array(
            self._df_predict(
                pd.DataFrame(data=np.array(processed_data), columns=sorted(raster_data))
            ),
            dtype="float64",
        )
        processed_numpy = processed_numpy.reshape([max_raster_rows, max_raster_columns])
        processed_raster = arcpy.NumPyArrayToRaster(
            processed_numpy,
            arcpy.Point(xmin, ymin),
            x_cell_size=min_cell_size_x,
            y_cell_size=min_cell_size_y,
        )
        processed_raster.save(output_folder_path)

        return True

    def show_results(self, rows=5):
        """
        Prints the rows of the dataframe with target and prediction columns.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer.
                                Number of rows to print.
        =====================   ===========================================

        :return: dataframe
        """
        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")
        self._check_requisites()
        min_size = len(self._data._validation_indexes)
        if min_size > rows:
            min_size = rows

        sample_indexes = random.sample(self._data._validation_indexes, min_size)
        if self._data._is_classification:
            rows_df = self._data._dataframe.loc[sample_indexes]
        else:
            rows_df = self._data._dataframe.iloc[sample_indexes]
        predictions = self._df_predict(rows_df)
        pd.options.mode.chained_assignment = None
        rows_df["prediction_results"] = predictions

        return rows_df

    def score(self):
        """
        :return: R2 score for regression model and Accuracy for classification model.
        """

        self._check_requisites()
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        # using loc instead of iloc to get data when dataframe doesnt have continuous indexes
        if self._data._is_classification:
            validation_dataframe = self._data._dataframe.loc[
                self._data._validation_indexes
            ].reset_index(drop=True)
        else:
            validation_dataframe = self._data._dataframe.iloc[
                self._data._validation_indexes
            ].reset_index(drop=True)

        predictions = np.array(self._df_predict(validation_dataframe))
        labels = validation_dataframe[self._data._dependent_variable]

        if self._data._is_classification:
            return (np.array(predictions) == labels).mean()
        else:
            return float(
                r2_score(
                    torch.tensor(np.array(predictions, dtype="float64")),
                    torch.tensor(np.array(labels, dtype="float64")),
                )
            )



import json
import arcgis
import torch
import types
import pickle
import warnings
import numpy as np
import pandas as pd
import datetime as dt
from torch import Tensor
from pathlib import Path
from torch.utils import data
from torch.utils.data import DataLoader
import mimetypes, os, random, sys, math
from fastai.data_block import DataBunch
from collections import Counter
from .._data import _prepare_working_dir
import matplotlib.pylab as plt
from fastai.data_block import get_files
from fastprogress.fastprogress import progress_bar
from .._utils.common import get_top_padding, ArcGISMSImage


def get_device():
    if getattr(arcgis.env, "_processorType", "") == "GPU" and torch.cuda.is_available():
        device = torch.device("cuda")
    elif getattr(arcgis.env, "_processorType", "") == "CPU":
        device = torch.device("cpu")
    else:
        device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
    return device


def ts_normalization(x, m, s):
    x = x.numpy()
    x = np.rollaxis(x, 2)  # TxCxS -> SxTxC
    x = (x - m) / s
    x = np.swapaxes((np.rollaxis(x, 1)), 1, 2)
    return torch.tensor(x)


def band_adjust(index, x0):
    band_index = np.array(index) - 1
    x0 = np.concatenate(
        [x0[:, q, :][:, None, :] for q in range(x0.shape[1]) if q in band_index],
        axis=1,
    )
    return x0


def time_adjust(index, x0):
    time_index = np.array(index) - 1
    x0 = np.concatenate(
        [x0[q, :, :][None, :, :] for q in range(x0.shape[0]) if q in time_index],
        axis=0,
    )
    return x0


class PSATAEDataset(data.Dataset):
    def __init__(
        self,
        folder,
        labels,
        npixel,
        sub_classes=None,
        norm=None,
        extra_feature=None,
        jitter=(0.01, 0.05),
        return_id=False,
        use_band_index=None,
        use_time_index=None,
    ):
        """
        Args:
            folder (str): path to the main folder of the dataset, formatted as indicated in the readme
            labels (str): name of the nomenclature to use in the labels.json file
            npixel (int): Number of sampled pixels in each parcel
            sub_classes (list): If provided, only the samples from the given list of classes are considered.
            (Can be used to remove classes with too few samples)
            norm (tuple): (mean,std) tuple to use for normalization
            extra_feature (str): name of the additional static feature file to use
            jitter (tuple): if provided (sigma, clip) values for the addition random gaussian noise
            return_id (bool): if True, the id of the yielded item is also returned (useful for inference)
        """
        super(PSATAEDataset, self).__init__()

        self.folder = folder
        if labels == "train_labels":
            self.data_folder = os.path.join(folder, "DATA", "train_data")
        else:
            self.data_folder = os.path.join(folder, "DATA", "valid_data")
        self.meta_folder = os.path.join(folder, "META")
        self.labels = labels
        self.npixel = npixel
        self.norm = norm

        self.use_band_index = use_band_index
        self.use_time_index = use_time_index

        self.extra_feature = extra_feature
        self.jitter = jitter  # (sigma , clip )
        self.return_id = return_id

        l = [f for f in os.listdir(self.data_folder) if f.endswith(".npy")]
        self.pid = [int(f.split(".")[0]) for f in l]
        self.pid = list(np.sort(self.pid))

        self.pid = list(map(str, self.pid))
        self.len = len(self.pid)

        # Get Labels
        if sub_classes is not None:
            sub_indices = []
            num_classes = len(sub_classes)
            convert = dict((c, i) for i, c in enumerate(sub_classes))

        with open(os.path.join(folder, "META", "labels.json"), "r") as file:
            d = json.loads(file.read())
            self.target = []
            for i, p in enumerate(self.pid):
                t = d[labels][str(int(p) - 1)]
                self.target.append(t)
                if sub_classes is not None:
                    if t in sub_classes:
                        sub_indices.append(i)
                        self.target[-1] = convert[self.target[-1]]
        if sub_classes is not None:
            self.pid = list(np.array(self.pid)[sub_indices])
            self.target = list(np.array(self.target)[sub_indices])
            self.len = len(sub_indices)

        if os.path.exists(os.path.join(folder, "META", "dates.json")):
            with open(os.path.join(folder, "META", "dates.json"), "r") as file:
                d = json.loads(file.read())
            self.dates = [d[str(i)] for i in range(len(d))]
            self.date_positions = date_positions(self.dates)
        else:
            self.date_positions = None

        if self.extra_feature is not None:
            with open(
                os.path.join(self.meta_folder, "{}.json".format(extra_feature)), "r"
            ) as file:
                self.extra = json.loads(file.read())

            if isinstance(self.extra[list(self.extra.keys())[0]], int):
                for k in self.extra.keys():
                    self.extra[k] = [self.extra[k]]
            df = pd.DataFrame(self.extra).transpose()
            self.extra_m, self.extra_s = (
                np.array(df.mean(axis=0)),
                np.array(df.std(axis=0)),
            )

    def __len__(self):
        return self.len

    def __getitem__(self, item):
        """
        Returns a Pixel-Set sequence tensor with its pixel mask and optional additional features.
        For each item npixel pixels are randomly dranw from the available pixels.
        If the total number of pixel is too small one arbitrary pixel is repeated. The pixel mask keeps track of true
        and repeated pixels.
        Returns:
              (Pixel-Set, Pixel-Mask) with:
              Pixel-Set: Sequence_length x Channels x npixel
              Pixel-Mask : Sequence_length x npixel
        """
        if self.labels == "train_labels":
            x0 = np.load(
                os.path.join(
                    self.folder, "DATA", "train_data", "{}.npy".format(self.pid[item])
                )
            )
            if self.use_band_index:
                x0 = band_adjust(self.use_band_index, x0)
            if self.use_time_index:
                x0 = time_adjust(self.use_time_index, x0)
        else:
            x0 = np.load(
                os.path.join(
                    self.folder, "DATA", "valid_data", "{}.npy".format(self.pid[item])
                )
            )
            if self.use_band_index:
                x0 = band_adjust(self.use_band_index, x0)
            if self.use_time_index:
                x0 = time_adjust(self.use_time_index, x0)

        y = self.target[item]

        if x0.shape[-1] > self.npixel:
            idx = np.random.choice(
                list(range(x0.shape[-1])), size=self.npixel, replace=False
            )
            x = x0[:, :, idx]
            mask = np.ones(self.npixel)

        elif x0.shape[-1] < self.npixel:
            if x0.shape[-1] == 0:
                x = np.zeros((*x0.shape[:2], self.npixel))
                mask = np.zeros(self.npixel)
                mask[0] = 1
            else:
                x = np.zeros((*x0.shape[:2], self.npixel))
                x[:, :, : x0.shape[-1]] = x0
                x[:, :, x0.shape[-1] :] = np.stack(
                    [x[:, :, 0] for _ in range(x0.shape[-1], x.shape[-1])], axis=-1
                )
                mask = np.array(
                    [1 for _ in range(x0.shape[-1])]
                    + [0 for _ in range(x0.shape[-1], self.npixel)]
                )
        else:
            x = x0
            mask = np.ones(self.npixel)

        if self.norm is not None:
            m, s = self.norm
            m = np.array(m)
            s = np.array(s)

            if len(m.shape) == 0:
                x = (x - m) / s
            elif len(m.shape) == 1:  # Normalise channel-wise
                x = (x.swapaxes(1, 2) - m) / s
                x = x.swapaxes(1, 2)  # Normalise channel-wise for each date
            elif len(m.shape) == 2:
                x = np.rollaxis(x, 2)  # TxCxS -> SxTxC
                x = (x - m) / s
                x = np.swapaxes((np.rollaxis(x, 1)), 1, 2)
        x = x.astype("float")

        if self.jitter is not None:
            sigma, clip = self.jitter
            x = x + np.clip(sigma * np.random.randn(*x.shape), -1 * clip, clip)

        mask = np.stack(
            [mask for _ in range(x.shape[0])], axis=0
        )  # Add temporal dimension to mask
        data = (Tensor(x), Tensor(mask))

        if self.extra_feature is not None:
            ef = (self.extra[str(self.pid[item])] - self.extra_m) / self.extra_s
            ef = torch.from_numpy(ef).float()

            ef = torch.stack([ef for _ in range(data[0].shape[0])], dim=0)
            data = (data, ef)

        if self.return_id:
            return data, torch.from_numpy(np.array(y, dtype=int)), self.pid[item]
        else:
            return data, torch.from_numpy(np.array(y, dtype=int))

    def __repr__(self):
        item = self.__getitem__(0)
        return (
            f"{self.__class__.__name__}, Tensor:{(item[0][0].shape)}, items:{self.len }"
        )


def parse(date):
    d = str(date)
    return int(d[:4]), int(d[4:6]), int(d[6:])


def interval_days(date1, date2):
    return abs((dt.datetime(*parse(date1)) - dt.datetime(*parse(date2))).days)


def date_positions(dates):
    pos = []
    for d in dates:
        pos.append(interval_days(d, dates[0]))
    return pos


image_extensions = set(
    k for k, v in mimetypes.types_map.items() if v.startswith("image/")
)


def create_train_val_sets(path, val_split_pct, working_dir, **kwargs):
    path = Path(path)
    images, labels = os.path.join(path, "images"), os.path.join(path, "labels")
    label_folds = [file for file in os.listdir(labels)]

    empty_folds = [
        i for i in label_folds if len(os.listdir(os.path.join(labels, i))) == 0
    ]
    if empty_folds:
        warnings.warn("Warning: Empty folder of label %s" % empty_folds)
        raise Exception()

    label_lst = get_files(labels, extensions=image_extensions, recurse=True)

    if working_dir is not None:
        save_path = working_dir
    else:
        save_path = path

    isExist_data = os.path.exists(os.path.join(save_path, "DATA"))
    isExist_train_data = os.path.exists(os.path.join(save_path, "DATA", "train_data"))
    isExist_valid_data = os.path.exists(os.path.join(save_path, "DATA", "valid_data"))
    isExist_labels = os.path.exists(os.path.join(save_path, "META"))

    nt = kwargs.get("n_temporal")
    ntemp = int(nt) if nt else nt
    npixels = kwargs.get("min_points", 64)
    ntempdates = kwargs.get("n_temporal_dates")
    use_band_index = kwargs.get("channels_of_interest")
    use_time_index = kwargs.get("timesteps_of_interest")
    interest_class = kwargs.get("classes_of_interest")

    emd_path = os.path.join(path, "esri_model_definition.emd")
    with open(emd_path) as f:
        emd_stats = json.load(f)
    IsMultidimensional = emd_stats.get("IsMultidimensional", False)
    serialDates = emd_stats.get("DimensionValues")
    imagespace = emd_stats.get("ImageSpaceUsed")

    if IsMultidimensional:
        ntempdates = [
            str(
                (
                    dt.datetime.fromordinal(dt.datetime(1900, 1, 1).toordinal() + i - 2)
                ).date()
            )
            for i in serialDates
        ]
        ntemp = len(ntempdates)

    if not interest_class:
        interest_class = [int(x) for x in label_folds]

    class_mapping_dict = dict()
    for i, n in enumerate(label_folds):
        if int(n) in interest_class:
            class_mapping_dict[i] = int(n)

    if interest_class:
        convertmap = dict((c, i) for i, c in enumerate(list(class_mapping_dict.keys())))
        convertmap = {v: k for k, v in convertmap.items()}
    if (
        not isExist_data
        or not isExist_train_data
        or not isExist_valid_data
        or not isExist_labels
    ):
        if not isExist_data:
            os.makedirs(os.path.join(save_path, "DATA"))
            if not isExist_train_data:
                os.makedirs(os.path.join(save_path, "DATA", "train_data"))
            if not isExist_valid_data:
                os.makedirs(os.path.join(save_path, "DATA", "valid_data"))
        if not isExist_labels:
            os.makedirs(os.path.join(save_path, "META"))

        train_labels, valid_labels = [], []

        for i in label_folds:
            file_lst = get_files(
                os.path.join(save_path, labels, i),
                extensions=image_extensions,
                recurse=False,
            )
            num = len(file_lst)

            val_num_images = int(val_split_pct * num)

            train_labels.extend(file_lst[val_num_images:])
            valid_labels.extend(file_lst[:val_num_images])

        val_lab_app, train_lab_app, trainlabs, validlabs = [], [], [], []
        train_arrs_lst, valid_arrs_lst, whole_data = [], [], []
        cnt, u, cnt1, s, cnt_lst_t, cnt_lst_v = 0, 1, 0, 1, [], []

        for i in progress_bar(label_lst):
            lab_name = os.path.split(i)[1]
            class_name = int(os.path.split(os.path.split(i)[0])[1])
            img_file, lab_file = (
                ArcGISMSImage.open(os.path.join(images, lab_name)),
                ArcGISMSImage.open(i),
            )
            img_arr, lab_arr = (
                img_file.data,
                torch.sum(lab_file.data, axis=0)[None, :, :],
            )

            concate_arr = torch.cat((img_arr, lab_arr), axis=0)
            indxs = torch.where(concate_arr[-1, :, :] != 0)
            timeseries_arr = concate_arr[:, indxs[0], indxs[1]][:-1, :]

            nchannel = int(timeseries_arr.shape[0] / ntemp)
            final = torch.reshape(
                timeseries_arr, (ntemp, nchannel, timeseries_arr.shape[1])
            )

            lol1 = lambda lst, sz: [
                len(lst[i : i + sz]) for i in range(0, len(lst), sz)
            ]
            if i in train_labels:
                train_arrs_lst.append(final)
                trainlabs.append(torch.tensor(class_name))
                num_shp = list(np.arange(final.shape[2]))
                u = 1 + cnt
                cnt += len(lol1(num_shp, npixels))
                save_cnt_lst = list(np.arange(u, cnt + 1))
                cntt = len(list(np.arange(u, cnt + 1)))
                train_lab_app.extend(cntt * [class_name])
                cnt_lst_t.extend(save_cnt_lst)

                def save(p, x, j, save_cnt):
                    l = np.save(
                        os.path.join(p, "DATA", "train_data", str(save_cnt[j])),
                        x.numpy(),
                    )
                    return x

                lol2 = lambda lst, sz: [
                    save(save_path, lst[:, :, i : i + sz], j, save_cnt_lst)
                    for j, i in enumerate(range(0, lst.shape[2], sz))
                ]
                divided = lol2(final, npixels)
            else:
                valid_arrs_lst.append(final.numpy())
                validlabs.append(class_name)
                num_shp = list(np.arange(final.shape[2]))
                s = 1 + cnt1
                cnt1 += len(lol1(num_shp, npixels))
                save_cnt_lst = list(np.arange(s, cnt1 + 1))
                cntt = len(list(np.arange(s, cnt1 + 1)))
                val_lab_app.extend(cntt * [class_name])
                cnt_lst_v.extend(save_cnt_lst)

                def save(p, x, j, save_cnt):
                    l = np.save(
                        os.path.join(p, "DATA", "valid_data", str(save_cnt[j])),
                        x.numpy(),
                    )
                    return x

                lol2 = lambda lst, sz: [
                    save(save_path, lst[:, :, i : i + sz], j, save_cnt_lst)
                    for j, i in enumerate(range(0, lst.shape[2], sz))
                ]
                divided = lol2(final, npixels)
            whole_data.append(final.numpy())

        stats_std_mean = np.concatenate(whole_data, axis=2)
        mean_std = (np.mean(stats_std_mean, axis=2), np.std(stats_std_mean, axis=2))

        with open(
            os.path.join(save_path, "META", "train_arrs_with_label.pickle"), "wb"
        ) as f:
            pickle.dump((train_arrs_lst, trainlabs), f)
        with open(os.path.join(save_path, "META", "mean_std.pickle"), "wb") as f:
            pickle.dump(mean_std, f)

        labs_dict = {
            "train": train_lab_app,
            "cn_t": len(cnt_lst_t),
            "valid": val_lab_app,
            "cn_v": len(cnt_lst_v),
        }
        labels_dic = {}

        for k in labs_dict.keys():
            if k in ["train", "valid"]:
                items = list(Counter(labs_dict[k]).keys())
                lab = dict()
                for i, j in zip(items, range(len(items))):
                    lab[str(i)] = int(j)
                re_lst = [
                    lab[str(z)] for z in labs_dict[k] if str(z) in list(lab.keys())
                ]
            if k == "train":
                Id_with_train_labels = dict()
                for i, j in zip(range(labs_dict["cn_t"]), re_lst):
                    Id_with_train_labels[str(i)] = int(j)
                labels_dic["train_labels"] = Id_with_train_labels
            else:
                Id_with_valid_labels = dict()
                for i, j in zip(range(labs_dict["cn_v"]), re_lst):
                    Id_with_valid_labels[str(i)] = int(j)
                labels_dic["valid_labels"] = Id_with_valid_labels

        labels_dic["npixel"] = int(npixels)
        with open(os.path.join(save_path, "META", "labels.json"), "w") as file:
            file.write(json.dumps(labels_dic, indent=4))
        if ntempdates:
            if use_time_index:
                timeidx = np.array(ntempdates) - 1
                ntempdates = ntempdates[timeidx]
            datelist = [x.replace("-", "") for x in ntempdates]
            dates_dict = {}
            for i in range(len(datelist)):
                dates_dict[str(i)] = datelist[i]
            with open(os.path.join(save_path, "META", "dates.json"), "w") as file:
                file.write(json.dumps(dates_dict, indent=4))

    if not [
        i
        for i in os.listdir(os.path.join(save_path, "META"))
        if i in ["labels.json", "mean_std.pickle", "train_arrs_with_label.pickle"]
    ]:
        warnings.warn(
            "required files missing, remove DATA and META folders at %s"
            % os.path.join(save_path)
        )
        raise Exception()

    mean_std = list(
        pickle.load(open(os.path.join(save_path, "META", "mean_std.pickle"), "rb"))
    )

    if use_band_index:
        band_index = np.array(use_band_index) - 1
        mean_std[0] = np.concatenate(
            [
                mean_std[0][:, q][:, None]
                for q in range(mean_std[0].shape[1])
                if q in band_index
            ],
            axis=1,
        )
        mean_std[1] = np.concatenate(
            [
                mean_std[1][:, q][:, None]
                for q in range(mean_std[1].shape[1])
                if q in band_index
            ],
            axis=1,
        )
    if use_time_index:
        time_index = np.array(use_time_index) - 1
        mean_std[0] = np.concatenate(
            [
                mean_std[0][q, :][None, :]
                for q in range(mean_std[0].shape[0])
                if q in time_index
            ],
            axis=0,
        )
        mean_std[1] = np.concatenate(
            [
                mean_std[1][q, :][None, :]
                for q in range(mean_std[1].shape[0])
                if q in time_index
            ],
            axis=0,
        )

    with open(os.path.join(save_path, "META", "labels.json"), "r") as file:
        lab_load = json.loads(file.read())
    train_arrs_lst, labs = pickle.load(
        open(os.path.join(save_path, "META", "train_arrs_with_label.pickle"), "rb")
    )

    timestep_infer, channels_infer = (
        train_arrs_lst[0].shape[0],
        train_arrs_lst[0].shape[1],
    )

    npixels = lab_load["npixel"]

    train_dataset = PSATAEDataset(
        save_path,
        labels="train_labels",
        npixel=npixels,
        norm=tuple(mean_std),
        sub_classes=list(class_mapping_dict.keys()),
        use_band_index=use_band_index,
        use_time_index=use_time_index,
    )
    valid_dataset = PSATAEDataset(
        save_path,
        labels="valid_labels",
        npixel=npixels,
        norm=tuple(mean_std),
        sub_classes=list(class_mapping_dict.keys()),
        use_band_index=use_band_index,
        use_time_index=use_time_index,
    )

    new_ntemp = None
    if use_time_index:
        new_ntemp = len(use_time_index)  # train_arrs_lst[0].shape[0]
    nt = new_ntemp if new_ntemp else ntemp

    return [
        (train_dataset, valid_dataset),
        class_mapping_dict,
        mean_std,
        train_arrs_lst,
        labs,
        npixels,
        nt,
        IsMultidimensional,
        imagespace,
        use_band_index,
        use_time_index,
        convertmap,
        timestep_infer,
        channels_infer,
    ]


def create_dataloaders(datasets, batch_size, dataloader_kwargs):
    dl_list = []
    for c, d in enumerate(datasets):
        if c == 0:
            dataloader_kwargs["shuffle"] = True
        else:
            dataloader_kwargs["shuffle"] = True
        dl = DataLoader(d, batch_size, **dataloader_kwargs)
        dl_list.append(dl)
    return dl_list


def prepare_psetae_data(
    path,
    batch_size,
    val_split_pct,
    working_dir,
    class_mapping,
    **kwargs,
):
    train_val_dataset = create_train_val_sets(
        path, val_split_pct, working_dir, **kwargs
    )
    databunch_kwargs = (
        {"num_workers": 0, "drop_last": True}
        if sys.platform == "win32"
        else {"num_workers": os.cpu_count() - 4, "drop_last": True}
    )
    train_dl, valid_dl = create_dataloaders(
        train_val_dataset[0], batch_size, databunch_kwargs
    )
    device = get_device()
    data = DataBunch(train_dl, valid_dl, device=device)
    if working_dir is not None:
        path = Path(os.path.abspath(working_dir))
    data.path = Path(path)
    data._temp_folder = _prepare_working_dir(path)
    data._class_map_dict = train_val_dataset[1].copy()
    data._num_class_map_dict = train_val_dataset[1].copy() if class_mapping else None

    if class_mapping:
        for r in data._class_map_dict.keys():
            data._class_map_dict[r] = class_mapping[data._class_map_dict[r]]

    data.show_batch = types.MethodType(show_batch, data)
    data._dataset_type = "PSETAE"
    data._date_positions = train_dl.date_positions
    data._n_temp = train_val_dataset[6]
    data._n_channel = data.train_ds[0][0][0].shape[1]
    data._mean_norm_stats = train_val_dataset[2][0]
    data._std_norm_stats = train_val_dataset[2][1]
    data._train_arrs_lst = train_val_dataset[3]
    data._train_labs = train_val_dataset[4]
    data._npixel = train_val_dataset[5]
    data._is_multidimensional = train_val_dataset[7]
    data._imagespace = train_val_dataset[8]
    data._bandindex = train_val_dataset[9]
    data._timeindex = train_val_dataset[10]
    data._convertmap = train_val_dataset[11]
    data._timestep_infer = train_val_dataset[12]
    data._channels_infer = train_val_dataset[13]

    return data


def show_batch(self, rows=5, spectral_view=False, **kwargs):
    """
    Show pixels temporal or spectral view of data in `ds_type`.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional int. number of pixels to be sampled
                            for each class.
    =====================   ===========================================
    spectral_view           Optional boolean. If set to True, Shows spectral
                            curves of sampled pixels.
    =====================   ===========================================
    **kwargs**

    """
    temporal_view = not spectral_view
    m, s = self._mean_norm_stats, self._std_norm_stats
    m = np.array(m)
    s = np.array(s)

    n = len(self._class_map_dict.keys())
    fig, axs = plt.subplots(
        math.ceil(n / 2), 2, squeeze=False, figsize=(20, math.ceil(n / 2) * 5)
    )
    top = get_top_padding(title_font_size=16, nrows=2, imsize=5)
    plt.subplots_adjust(top=top, hspace=0.3)

    viz_data_seg = dict((k, []) for k in self._class_map_dict.keys())
    class_dict = (
        self._num_class_map_dict if self._num_class_map_dict else self._class_map_dict
    )
    rev_map = {v: k for k, v in class_dict.items()}

    for t, r in zip(self._train_arrs_lst, self._train_labs):
        if int(r) in rev_map.keys():
            r = rev_map[int(r)]
            if self._bandindex:
                t = band_adjust(self._bandindex, t)
            if self._timeindex:
                t = time_adjust(self._timeindex, t)
            t = ts_normalization(Tensor(t), m, s)
            if t.shape[2] <= self._npixel:
                viz_data_seg[int(r)].append(t)
            else:
                index = np.random.choice(t.shape[2], self._npixel, replace=False)
                viz_data_seg[int(r)].append(t.index_select(2, torch.tensor(index)))

    for z in range(n):
        mp = list(self._class_map_dict.keys())
        class_1_arr = np.concatenate(viz_data_seg[mp[z]], axis=2)
        randlist = random.sample(list(range(class_1_arr.shape[2])), rows)
        if class_1_arr.shape[0] != 1:
            if temporal_view:
                bands = [("T" + str(x)) for x in range(class_1_arr.shape[0])]
                colours = plt.cm.jet(np.linspace(0, 1, class_1_arr.shape[1]))
                for j in range(class_1_arr.shape[1]):
                    for o, i in enumerate(randlist):
                        ax = axs.flat[z]
                        if o == 0:
                            lab = "band " + str(j + 1)
                        else:
                            lab = None
                        ax.plot(
                            bands,
                            class_1_arr[:, :, i][:, j],
                            color=colours[j],
                            label=lab,
                        )
                        ax.legend(loc="upper right")
            else:
                bands = [("B" + str(x)) for x in range(class_1_arr.shape[1])]
                colours = plt.cm.jet(np.linspace(0, 1, class_1_arr.shape[0]))
                for j in range(class_1_arr.shape[0]):
                    for o, i in enumerate(randlist):
                        ax = axs.flat[z]
                        if o == 0:
                            lab = "T " + str(j + 1)
                        else:
                            lab = None
                        ax.plot(
                            bands,
                            class_1_arr[:, :, i][j, :],
                            color=colours[j],
                            label=lab,
                        )
                        ax.legend(loc="upper right")
        else:
            bands = [("B" + str(x)) for x in range(class_1_arr.shape[1])]
            colours = plt.cm.jet(np.linspace(0, 1, class_1_arr.shape[0]))
            ax = axs.flat[z]
            lab = "T " + str(1)
            ax.plot(
                bands, class_1_arr[:, :, randlist][0, :], color=colours[0], label=lab
            )
            ax.legend(loc="upper right")
        ax.set_title(label=str(self._class_map_dict[mp[z]]), fontsize=17, pad=10)
    for ax in axs.flat:
        if not bool(ax.has_data()):
            fig.delaxes(ax)
    plt.show()


def show_results(self, rows, **kwargs):
    validarr = torch.cat([i[0][None, :, :, :] for i, j in self._data.valid_ds], axis=0)
    labsarr = torch.stack([j for i, j in self._data.valid_ds])

    final_img = torch.moveaxis(validarr, 3, 1)[:, :, :, :, None]
    img_arr = torch.reshape(
        final_img,
        (
            final_img.shape[0] * final_img.shape[1],
            final_img.shape[2],
            final_img.shape[3],
            1,
        ),
    ).to(self._device)
    final_labs = [
        j for l in range(validarr.shape[0]) for j in validarr.shape[3] * [labsarr[l]]
    ]
    final_labs = torch.stack(final_labs).to(self._device)
    divided = DataLoader(img_arr, batch_size=65536, pin_memory=False)

    prediction = []
    for i in divided:
        sim = torch.ones(i.shape[0], i.shape[1], 1).to(self._device)
        self.learn.model.eval()
        with torch.no_grad():
            pred = self.learn.model(i, sim)
        prediction.append(pred.argmax(dim=1))

    df = pd.DataFrame(
        np.stack(
            [torch.cat(prediction).cpu().numpy(), final_labs.cpu().numpy()], axis=1
        ),
        columns=["Predicted", "True"],
    )

    df.sort_values(by=["True"], ascending=True)
    clses = list(self._data._class_map_dict.keys())
    num_pixels = rows * 2
    grouped = df.groupby("True").apply(lambda x: x.sample(int(num_pixels / len(clses))))
    all_cls_arr = img_arr.index_select(
        0, torch.tensor([j for i, j in grouped["Predicted"].index]).to(self._device)
    )
    if grouped.shape[0] != num_pixels:
        remain = df.sample(abs(grouped.shape[0] - num_pixels))
        remained_all_cls_arr = img_arr.index_select(
            0, torch.tensor([i for i in remain.index]).to(self._device)
        )
        all_cls_arr = torch.cat((all_cls_arr, remained_all_cls_arr), axis=0)
    bands = [("T" + str(x)) for x in range(all_cls_arr.shape[1])]
    pred = (
        list(grouped["Predicted"].values) + [i for i in remain["Predicted"]]
        if grouped.shape[0] != num_pixels
        else grouped["Predicted"].values
    )
    true = (
        list(grouped["True"].values) + [i for i in remain["True"]]
        if grouped.shape[0] != num_pixels
        else grouped["True"].values
    )
    pred.sort(), true.sort()

    if self._data._convertmap:
        true = np.array([self._data._convertmap.get(item, item) for item in true])
        pred = np.array([self._data._convertmap.get(item, item) for item in pred])

    trues = np.array([self._data._class_map_dict.get(item, item) for item in true])
    preds = np.array([self._data._class_map_dict.get(item, item) for item in pred])
    n = all_cls_arr.shape[0]
    fig, axs = plt.subplots(
        math.ceil(n / 2), 2, squeeze=False, figsize=(20, math.ceil(n / 2) * 5)
    )
    top = get_top_padding(title_font_size=16, nrows=math.ceil(n / 2), imsize=5)
    plt.subplots_adjust(top=top, hspace=0.33)
    axs1 = np.reshape(axs, (axs.shape[0] * axs.shape[1]))
    for i in range(axs1.shape[0]):
        for k in range(all_cls_arr[i].shape[1]):
            axs1[i].plot(bands, all_cls_arr[i][:, k, :].cpu().numpy())
            if preds[i] == trues[i]:
                col = "blue"
            else:
                col = "red"
            if i == 0:
                axs1[0].text(
                    0.98,
                    1.4,
                    "Ground truth",
                    fontsize=17,
                    transform=axs1[0].transAxes,
                    fontweight="bold",
                )
                axs1[0].text(
                    1,
                    1.3,
                    "Predictions",
                    fontsize=17,
                    transform=axs1[0].transAxes,
                    fontweight="bold",
                )

            axs1[i].text(
                0.5,
                1.15,
                str(trues[i]),
                fontsize=17,
                transform=axs1[i].transAxes,
                color=col,
                ha="center",
                va="center",
            )
            axs1[i].text(
                0.5,
                1.05,
                str(preds[i]),
                fontsize=17,
                transform=axs1[i].transAxes,
                color=col,
                ha="center",
                va="center",
            )
    for ax in axs.flat:
        if not bool(ax.has_data()):
            fig.delaxes(ax)



import torch
import numpy as np
import math
import torch.nn.functional as F
from .._utils.env import is_arcgispronotebook
import warnings


def calculate_intersection(preds, targs, mode):
    """
    Calculates intersection between preds and targs.
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor.
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor.
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean'
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns intersection between preds and targs -> torch.tensor
    """

    if mode == "per_class":
        dim = 2
        assert targs.ndim == 3
    if mode == "mean":
        dim = 1
        assert targs.ndim == 2

    return (preds * targs).sum(dim=dim)


def calculate_union(preds, targs, intersection, mode):
    """
    Calculates union between preds and targs.
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor.
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor.
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    intersection            Required torch.tensor.
                            Intersection between preds and targs
                            calculate_intersection(preds, targs, mode)
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean'
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns union between preds and targs -> torch.tensor
    """
    if mode == "per_class":
        dim = 2
        assert targs.ndim == 3
    if mode == "mean":
        dim = 1
        assert targs.ndim == 2

    return (preds + targs).sum(dim=dim) - intersection


def confusion_matrix(preds, targs, intersection, mode):
    """
    Calculates true positive, false positive and false negetive
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor.
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor.
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean'
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns true positive, false positive and false negetive -> torch.tensor, torch.tensor, torch.tensor
    """

    if mode == "per_class":
        dim = 2
        assert targs.ndim == 3
    if mode == "mean":
        dim = 1
        assert targs.ndim == 2

    TP = intersection  # true positive
    FP = preds.sum(dim=dim) - intersection  # flase positive
    FN = targs.sum(dim=dim) - intersection  # false negetive

    return TP, FP, FN


def calculate_precision(TP, FP, eps=1e-08):
    return TP / (TP + FP + eps)


def calculate_recall(TP, FN, eps=1e-08):
    return TP / (TP + FN + eps)


def calculate_f1(precision, recall, eps=1e-08):
    return 2 * (precision * recall) / (precision + recall + eps)


def prepare_output_for_metrics(preds, targs, ignore_classes, mode):
    """
    This function modifies the shape of targs(ground truth segmentation mask)
    to match with preds(predicted output)
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor.
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor.
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    mode                    Required str.
                            Possible values = 'per_class' or 'mean'
                            'per_class' = Per class metrics are calculated.
                            'mean, = Mean metrics are calculated.
    -------------------------------------------------------------------
    Returns modeified preds and targs
    """

    from arcgis.learn._utils.segmentation_loss_functions import expand_outputs

    num = preds.size(0)
    classes = preds.size(1)
    keep_classes = [v for v in range(classes) if v not in ignore_classes]
    encoded_targs = expand_outputs(preds, targs)
    preds = preds.argmax(dim=1)
    preds = F.one_hot(preds, classes).permute(0, 3, 1, 2)
    if mode == "per_class":
        preds = preds.contiguous().float().view(num, classes, -1)[:, keep_classes, :]
        encoded_targs = encoded_targs.float().view(num, classes, -1)[:, keep_classes, :]
    if mode == "mean":
        preds = (
            preds.contiguous()
            .float()
            .view(num, classes, -1)[:, keep_classes, :]
            .view(num, -1)
        )
        encoded_targs = (
            encoded_targs.float()
            .view(num, classes, -1)[:, keep_classes, :]
            .view(num, -1)
        )
    return preds, encoded_targs


def calculate_metrics(preds, targs, mode, ignore_classes=[]):
    """
    Calculates precision, recall and f1 scores.
    """
    preds, targs = prepare_output_for_metrics(preds, targs, ignore_classes, mode)
    intersection = calculate_intersection(preds, targs, mode)
    # union = calculate_union(preds, targs, intersection, mode)
    tp, fp, fn = confusion_matrix(preds, targs, intersection, mode)
    precision = calculate_precision(tp, fp, eps=1e-08)
    recall = calculate_recall(tp, fn, eps=1e-08)
    f1 = calculate_f1(precision, recall, eps=1e-08)
    return precision, recall, f1


def post_process_CD(preds, y, n_classes):
    from .change_detection_data import post_process, post_process_y

    preds = post_process(preds)
    y[y == 1] = 0
    y[y == -1] = 1
    preds = preds.squeeze(1)
    preds = F.one_hot(preds, n_classes).permute(0, 3, 1, 2)
    return preds, y


def per_class_metrics(self, ignore_classes=[], **kwargs):
    """
    Computes per class precision, recall and f1 scores
    """
    mode = "per_class"
    import pandas as pd

    dl = kwargs.get("dl", None)
    postprocess_type = kwargs.get("postprocess_type", None)
    ignore_mapped_class = kwargs.get("ignore_mapped_class", ignore_classes)
    keep_classes = [
        c for i, c in enumerate(self._data.classes) if i not in ignore_mapped_class
    ]
    model = self.learn.model.eval()
    valid_class_len = len(self._data.classes) - len(ignore_classes)
    tp_counts = torch.zeros(valid_class_len)
    fp_counts = torch.zeros(valid_class_len)
    fn_counts = torch.zeros(valid_class_len)
    observed_val_data_classes = set()
    for batch in self._data.valid_dl if dl is None else dl:
        x, y = batch
        y = y.to("cpu")
        observed_val_data_classes = observed_val_data_classes.union(
            np.unique(y).tolist()
        )
        with torch.no_grad():
            if getattr(self, "_is_model_extension", False):
                if self._is_multispectral:
                    predictions = model(
                        self._model_conf.transform_input_multispectral(x)
                    ).to("cpu")
                else:
                    predictions = model(self._model_conf.transform_input(x)).to("cpu")
            elif type(x) is list or type(x) is tuple:
                predictions = model(*x).detach().to("cpu")
            else:
                predictions = model(x).detach().to("cpu")
            if postprocess_type == "CD":
                predictions, y = post_process_CD(
                    predictions, y, n_classes=len(self._data.classes)
                )
            predictions, y = prepare_output_for_metrics(
                predictions, y, ignore_classes, mode
            )
            intersection = calculate_intersection(predictions, y, mode)
            tp, fp, fn = confusion_matrix(predictions, y, intersection, mode)
            tp_counts = tp_counts + tp.sum(0)
            fp_counts = fp_counts + fp.sum(0)
            fn_counts = fn_counts + fn.sum(0)
    train_classes = set(self._data.class_mapping.keys()) - {0}
    observed_val_data_classes = observed_val_data_classes - {0}
    if len(train_classes) != len(observed_val_data_classes):
        warnings.warn(
            f'Validation dataset classes {sorted(list(observed_val_data_classes))} does not match the training dataset \
classes {sorted(list(train_classes))}, you could use "stratify=True" with prepare_data. If training classes are still \
missing in validation data you could try increasing the minority class samples. Metrics are only being calculated for \
classes present in the validation dataset.'
        )
    precision = calculate_precision(tp_counts, fp_counts)
    recall = calculate_recall(tp_counts, fn_counts)
    f1 = calculate_f1(precision, recall)
    columns = keep_classes
    df = pd.DataFrame(columns=columns)
    df.loc["precision"] = precision.tolist()
    df.loc["recall"] = recall.tolist()
    df.loc["f1"] = f1.tolist()
    return df


def show_batch_classified_tiles(self, rows=3, alpha=0.7, **kwargs):
    """
    This function randomly picks a few training chips and visualized them.
    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    rows                    Optional Integer.
                            Number of rows to display.
                            Default: 3.
    ---------------------   -------------------------------------------
    alpha                   Optional Float.
                            Opacity of the lables for the corresponding
                            images. Values range between 0 and 1, where
                            1 means opaque.
    -------------------------------------------------------------------
    """

    import matplotlib.pyplot as plt
    from .._utils.common import (
        kwarg_fill_none,
        get_nbatches,
        find_data_loader,
        get_top_padding,
        get_symbology_bands,
        dynamic_range_adjustment,
        denorm_x,
        image_tensor_checks_plotting,
    )

    imsize = kwarg_fill_none(kwargs, "imsize", 5)
    nrows = rows
    ncols = kwarg_fill_none(kwargs, "ncols", 3)
    imsize = kwarg_fill_none(kwargs, "imsize", 5)
    statistics_type = kwarg_fill_none(
        kwargs, "statistics_type", "dataset"
    )  # Accepted Values `dataset`, `DRA`

    n_items = kwargs.get("n_items", nrows * ncols)
    n_items = min(n_items, len(self.x))
    nrows = math.ceil(n_items / ncols)

    top = kwargs.get("top", None)
    title_font_size = 16
    if top is None:
        top = get_top_padding(
            title_font_size=title_font_size, nrows=nrows, imsize=imsize
        )

    # Get n batches
    type_data_loader = kwarg_fill_none(
        kwargs, "data_loader", "training"
    )  # options : traininig, validation, testing
    data_loader = find_data_loader(type_data_loader, self)
    x_batch, y_batch = get_nbatches(data_loader, math.ceil(n_items / self.batch_size))
    symbology_x_batch = x_batch = torch.cat(x_batch)
    y_batch = torch.cat(y_batch)

    symbology_bands = [0, 1, 2]
    if self._is_multispectral:
        # Get RGB Bands for plotting
        rgb_bands = kwarg_fill_none(kwargs, "rgb_bands", self._symbology_rgb_bands)

        # Get Symbology bands
        symbology_bands = get_symbology_bands(
            rgb_bands, self._extract_bands, self._bands
        )

    # Denormalize X
    x_batch = denorm_x(x_batch, self)

    # Extract RGB Bands for plotting
    symbology_x_batch = x_batch[:, symbology_bands]

    # Apply Image Strecthing
    if statistics_type == "DRA":
        symbology_x_batch = dynamic_range_adjustment(symbology_x_batch)

    symbology_x_batch = image_tensor_checks_plotting(symbology_x_batch)

    # Get color Array
    color_array = self._multispectral_color_array
    color_array[1:, 3] = alpha

    # Plot now
    fig, axs = plt.subplots(
        nrows=nrows, ncols=ncols, figsize=(ncols * imsize, nrows * imsize)
    )
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            axi = axs
            if nrows == 1:
                axi = axi
            else:
                axi = axi[r]
            if ncols == 1:
                axi = axi
            else:
                axi = axi[c]
            axi.axis("off")
            if idx < symbology_x_batch.shape[0]:
                axi.imshow(symbology_x_batch[idx].cpu().numpy())
                y_rgb = color_array[y_batch[idx][0]].cpu().numpy()
                axi.imshow(y_rgb, alpha=alpha)
            idx += 1
    if is_arcgispronotebook():
        plt.show()



from typing import Union, Optional, ClassVar
from dataclasses import field, dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.geometry import (
    _HasGeometry,
    SingleFieldGeometry,
    XYZGeometry,
)
from arcgis.realtime.velocity.feeds.time import _HasTime, TimeInstant, TimeInterval
from arcgis.realtime.velocity.input.format import (
    EsriJsonFormat,
    GeoJsonFormat,
    JsonFormat,
    DelimitedFormat,
    XMLFormat,
    _format_from_config,
)


@dataclass
class AWSIoT(_FeedTemplate, _HasTime, _HasGeometry):
    """
    Receive events from an AWS IoT broker. This data class can be used to define the feed configuration and
    to create the feed.

    ==================      ============================================================================================
    **Parameter**            **Description**
    ------------------      --------------------------------------------------------------------------------------------
    label                   String. Unique label for the feed instance.
    ------------------      --------------------------------------------------------------------------------------------
    description             String. Feed description.
    ------------------      --------------------------------------------------------------------------------------------
    endpoint                String. Endpoint for the AWS IoT broker.
    ------------------      --------------------------------------------------------------------------------------------
    topic                   String. Topic over which event messages stream.
    ------------------      --------------------------------------------------------------------------------------------
    qos_level               int. The Quality of Service (QoS) level defines the guarantee of delivery for a specific
                            message. A QoS of 0 means a message is delivered zero or more times. It offers better
                            performance, but no guaranteed delivery. A QoS of 1 means a message is delivered at least
                            once, thereby offering guaranteed delivery. With both levels, messages may be delivered
                            multiple times. The default is: 0.
    ==================      ============================================================================================

    =====================   ========================================================================================
    **Optional Argument**   **Description**
    =====================   ========================================================================================
    access_key_id           String. Access key ID for the AWS IoT credentials.
    ---------------------   ----------------------------------------------------------------------------------------
    secret_access_key       String. Secret access key for the AWS IoT credentials.
    ---------------------   ----------------------------------------------------------------------------------------
    session_token           String. Session token for the AWS IoT broker.
    ---------------------   ----------------------------------------------------------------------------------------
    data_format             [:class:`~arcgis.realtime.velocity.input.EsriJsonFormat`, :class:`~arcgis.realtime.velocity.input.GeoJsonFormat`, :class:`~arcgis.realtime.velocity.input.DelimitedFormat`, :class:`~arcgis.realtime.velocity.input.JsonFormat`, :class:`~arcgis.realtime.velocity.input.XMLFormat`].
                            An instance that contains the data format
                            configuration for this feed. Configure only allowed formats.
                            If this is not set right during initialization, a format will be
                            auto-detected and set from a sample of the incoming data. This sample
                            will be fetched from the configuration provided so far in the init.
    ---------------------   ----------------------------------------------------------------------------------------
    track_id_field          String. Name of the field from the incoming data that should be set as the
                            track ID.
    ---------------------   ----------------------------------------------------------------------------------------
    geometry                [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`]. An instance of geometry configuration
                            that will be used to create geometry objects from the incoming data.
    ---------------------   ----------------------------------------------------------------------------------------
    time                    [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                            will be used to create time information from the incoming data.
    =====================   ========================================================================================

    :return: A data class with AWS Iot feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import AWSIoT
        from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant

        aws_config = AWSIoT(
            label="feed_name",
            description="feed_description",
            endpoint="aws_iot feed endpoint",
            topic="aws_iot_topic",
            qos_level=0,
            access_key_id="aws_iot_access_key_id",
            secret_access_key="aws_iot_secret_access_key",
            data_format=None
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        aws_feed = feeds.create(aws_config)
        aws_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # AWS IoT specific properties
    endpoint: str
    topic: str
    qos_level: int = field(default=0)
    access_key_id: Optional[str] = None
    secret_access_key: Optional[str] = None
    session_token: Optional[str] = None

    # user can define these properties even after initialization
    data_format: Optional[
        Union[DelimitedFormat, EsriJsonFormat, GeoJsonFormat, JsonFormat, XMLFormat]
    ] = None
    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasGeometry properties
    geometry: Optional[Union[XYZGeometry, SingleFieldGeometry]] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None

    # FeedTemplate properties
    _name: ClassVar[str] = "awsiot"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        test_connection = self._util.test_connection(
            input_type="feed", payload=feed_properties
        )
        # Test connection to make sure feed can fetch schema
        if test_connection is True:
            # test connection succeeded. Now try getting sample messages
            sample_payload = {
                "properties": {
                    "maxSamplesToCollect": 5,
                    "timeoutInMillis": 5000,
                }
            }
            self._dict_deep_merge(sample_payload, feed_properties)
            # Sample messages to fetch schema/fields
            sample_messages = self._util.sample_messages(
                input_type="feed", payload=sample_payload
            )

            if sample_messages["featureSchema"] is not None:
                # sample messages succeeded. Get Feature Schema from it
                self._set_fields(sample_messages["featureSchema"])

            # if Format was not specified by user, use the auto-detected format from the sample messages response as this feed object's format.
            if self.data_format is None:
                self.data_format = _format_from_config(sample_messages)

            print(
                "Feature Schema retrieved from the Feed:",
                sample_messages["featureSchema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.geometry is not None:
                self.set_geometry_config(self.geometry)
            if self.time is not None:
                self.set_time_config(self.time)

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        if self.access_key_id:
            access_key_id_prop = {f"{self._name}.accessKeyId": self.access_key_id}
        else:
            access_key_id_prop = {}

        if self.secret_access_key:
            secret_access_key_prop = {
                f"{self._name}.secretAccessKey": self.secret_access_key
            }
        else:
            secret_access_key_prop = {}

        if self.session_token:
            session_token_prop = {f"{self._name}.sessionToken": self.session_token}
        else:
            session_token_prop = {}

        feed_properties = {
            "name": self._name,
            "properties": {
                f"{self._name}.endpoint": self.endpoint,
                f"{self._name}.topic": self.topic,
                f"{self._name}.qos": self.qos_level,
                **access_key_id_prop,
                **secret_access_key_prop,
                **session_token_prop,
            },
        }

        if self.data_format is not None:
            format_dict = self.data_format._build()
            self._dict_deep_merge(feed_properties, format_dict)

        return feed_properties



from typing import ClassVar, List


class _ReservedFields:
    """
    Field names that are reserved for internal use by Velocity. A user defined field name should not be one of these.
    """

    _reserved_names_gdb: ClassVar[List[str]] = [
        "ENTITY",
        "NUMPTS",
        "MINX",
        "MINY",
        "MAXX",
        "MAXY",
        "MINZ",
        "MAXZ",
        "MINM",
        "MAXM",
        "AREA",
        "LEN",
        "SRID",
        "POINTS",
        "Shape_Area",
        "DB2GSE.ST_Area(Shape)",
        "SE_Area(Shape)",
        "Shape.AREA",
        "st_area(Shape)",
        "Shape.area",
        "Shape.STArea()",
        "Shape_Length",
        "DB2GSE.SdeLength(Shape)",
        "SE_Length(Shape)",
        "Shape.LEN",
        "st_length(Shape)",
        "Shape.len",
        "Shape.STLength()",
        "Shape",
    ]
    _reserved_names_fs: ClassVar[List[str]] = [
        "fid",
        "area",
        "len",
        "ponts",
        "numofpts",
        "entity",
        "eminx",
        "eminy",
        "emaxx",
        "emaxy",
        "eminz",
        "emaxz",
        "min_measure",
        "max_measure",
    ]
    _reserved_names_service_query: ClassVar[List[str]] = [
        "shape__area",
        "shape__length",
    ]
    _reserved_names_es: ClassVar[List[str]] = ["_id"]
    _all_reserved_names_lower: ClassVar[List[str]] = list(
        map(
            str.lower,
            (
                _reserved_names_fs
                + _reserved_names_gdb
                + _reserved_names_service_query
                + _reserved_names_es
            ),
        )
    )

    @staticmethod
    def is_reserved(field_name: str) -> bool:
        """
        Checks if a given field name is a reserved name. This is a case-insensitive comparison.

        ==============     =====================================
        **Parameter**       **Description**
        --------------     -------------------------------------
        field_name         String. A field name to test
        ==============     =====================================

        :return: True if field_name is a reserved name
        """
        return field_name.lower() in _ReservedFields._all_reserved_names_lower

    @staticmethod
    def is_not_reserved(field_name: str) -> bool:
        """
        Checks if a given field name is not a reserved name. This is a case-insensitive comparison.

        ==============     =====================================
        **Parameter**       **Description**
        --------------     -------------------------------------
        field_name         String. A field name to test
        ==============     =====================================

        :return: True if field_name is not a reserved name
        """
        return not _ReservedFields.is_reserved(field_name)



from __future__ import annotations
from arcgis.gis import GIS
from arcgis._impl.common._isd import InsensitiveDict
from typing import Dict, Any, Union, Optional


class UsageStatistics:
    """
    Provides access to the metrics viewer and metrics API tools. As an
    administrator, you can use these tools to monitor GIS service usage in
    your organization for feature services, map services, tiled map
    services, geocode services, and geometry services. Information that
    can be gathered from service usage statistics include:

    - Historical service usage data
    - Peak and off-peak periods of service usage
    - Slowdown in response times or throughput



    """

    _gis = None
    _url = None
    _properties = None

    # ---------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS) -> None:
        """initializer"""
        self._url = url
        self._gis = gis

    # ---------------------------------------------------------------------
    @property
    def properties(self) -> InsensitiveDict:
        """returns the properties of the resource"""
        if self._properties is None:
            self._properties = InsensitiveDict(
                self._gis._con.get(self._url, {"f": "json"})
            )
        return self._properties

    # ---------------------------------------------------------------------
    def update_credentials(self, resource: str, username: str, password: str) -> bool:
        """
        Updates the credentials for the metrics viewer and metrics API.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        resource            Required String.  Specifies whether the updated credentials will be
                            applied to the metrics viewer (`grafana`) or the metrics
                            API (`prometheus`).

                            Values: `grafana` or `prometheus`
        ---------------     --------------------------------------------------------------------
        username            Required String. The new username for the specified metrics resource.
        ---------------     --------------------------------------------------------------------
        password            Required String. The new password for the metrics resource. The new
                            password must be a minimum of eight characters and must contain at
                            least one letter (A-Z, a-z), one number (0-9), and one special
                            character.
        ===============     ====================================================================

        :return: boolean
        """
        url = f"{self._url}/update/credentials"
        try:
            params = {
                "f": "json",
                "resource": resource,
                "username": username,
                "password": password,
            }
            resp = self._gis._con.post(url, params)
            return resp.get("status", "failed") == "success"
        except Exception as e:
            print(e)
            raise
        return False



from pathlib import Path
import json
import warnings
from ._model_extension import ModelExtension

try:
    from fastai.vision import flatten_model, ImageList
    from fastai.vision import imagenet_stats
    import torch
    from fastai.torch_core import split_model_idx
    from .._utils.pascal_voc_rectangles import ObjectDetectionCategoryList
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from ._arcgis_model import _resnet_family, _get_device

    HAS_FASTAI = True

except Exception as e:
    HAS_FASTAI = False


class CustomDetReg:
    """
    Create class with following fixed function names and the number of arguents to train your model from external source
    """

    import torch

    def get_model(self, data, backbone="resnet50", **kwargs):
        from arcgis.learn._utils import nested_tensor_from_tensor_list

        self.Nt = nested_tensor_from_tensor_list

        from arcgis.learn.models._detr_object_detection.backbone import (
            build_leran_backbone,
        )
        from arcgis.learn.models._detr_object_detection.deformable_transformer import (
            build_learn_deforamble_transformer,
        )
        from arcgis.learn.models._detr_object_detection.deformable_detr import (
            DeformableDETR,
            SetCriterion,
            PostProcess,
        )

        backbone = build_leran_backbone(backbone)
        transformer = build_learn_deforamble_transformer()

        model = DeformableDETR(
            backbone,
            transformer,
            num_classes=data.c,
            num_queries=100,
            num_feature_levels=4,
            aux_loss=True,
            with_box_refine=False,
            two_stage=False,
            object_embedding_loss=False,
            obj_embedding_head="intermediate",
        )

        self.model = model

        from arcgis.learn.models._detr_object_detection.matcher import HungarianMatcher

        matcher = HungarianMatcher(cost_class=2, cost_bbox=5, cost_giou=2)
        weight_dict = {"loss_ce": 2, "loss_bbox": 5, "loss_giou": 2}
        aux_weight_dict = {}
        dec_layers = 6
        for i in range(dec_layers - 1):
            aux_weight_dict.update({k + f"_{i}": v for k, v in weight_dict.items()})
        aux_weight_dict.update({k + f"_enc": v for k, v in weight_dict.items()})
        weight_dict.update(aux_weight_dict)
        losses = ["labels", "boxes", "cardinality"]

        self.criterion = SetCriterion(
            data.c, matcher, weight_dict, losses, focal_alpha=0.25
        ).to(data.device)

        self.postprocessors = PostProcess()

        # for resizing img during training
        self.scale_factor = 1.5

        return model

    def on_batch_begin(self, learn, model_input_batch, model_target_batch, **kwargs):
        target_list = []

        for bbox, label in zip(*model_target_batch):
            bbox = (bbox + 1) / 2  # change boxes in the range of 0-1.
            # convert to bboxes [xmin,ymin,xmax,ymax].
            bbox = self.torch.index_select(
                bbox,
                1,
                self.torch.tensor([1, 0, 3, 2]).to(learn.data.device),
            )
            mask = (bbox[:, 2:] > (bbox[:, :2])).all(1)
            bbox = bbox[mask]
            label = label[mask]

            def box_xyxy_to_cxcywh(x):
                x0, y0, x1, y1 = x.unbind(-1)
                b = [(x0 + x1) / 2, (y0 + y1) / 2, (x1 - x0), (y1 - y0)]
                return torch.stack(b, dim=-1)

            target = {}
            target["boxes"] = box_xyxy_to_cxcywh(bbox)
            target["labels"] = label
            target_list.append(target)

        # return model_input and model_target
        return (
            self.resize_input_batch(model_input_batch),
            target_list,
        )

    def resize_input_batch(self, input):
        input = self.torch.nn.functional.interpolate(
            input,
            scale_factor=self.scale_factor,
            mode="bilinear",
            recompute_scale_factor=True,
            align_corners=False,
        )

        return self.Nt(list(input))

    def transform_input(self, xb, thresh=0.5, nms_overlap=0.1):  # transform_input
        """
        function for feding the input to the model in validation/infrencing mode.

        xb - tensor with shape [N, C, H, W]
        """
        self.thres = thresh
        return self.resize_input_batch(xb)

    def transform_input_multispectral(self, xb, thresh=0.5, nms_overlap=0.1):
        self.thres = thresh
        return self.resize_input_batch(xb)

    def loss(self, model_output, *model_target):
        loss_dict = self.criterion(model_output, model_target)
        weight_dict = self.criterion.weight_dict
        losses = sum(
            loss_dict[k] * weight_dict[k] for k in loss_dict.keys() if k in weight_dict
        )
        return losses

    def post_process(
        self, pred, nms_overlap, thres, chip_size, device=torch.device("cuda")
    ):
        post_processed_pred = []
        pred_logits = pred["pred_logits"]
        results = self.postprocessors(
            pred,
            (self.torch.ones((pred_logits.shape[0], 2)) * chip_size).to(
                pred_logits.device
            ),
        )
        for result in results:
            bbox, label, score = result["boxes"], result["labels"], result["scores"]
            bbox = bbox / (chip_size / 2) - 1
            kip_pred = score > self.thres
            score = score[kip_pred]
            bbox = bbox[kip_pred]
            label = label[kip_pred]
            # convert bboxes in format [y1,x1,y2,x2]
            bbox = self.torch.index_select(
                bbox, 1, self.torch.tensor([1, 0, 3, 2]).to(bbox.device)
            )
            # Append the tuple in list for each image
            post_processed_pred.append(
                (bbox.data.to(device), label.to(device), score.to(device))
            )

        return post_processed_pred


class DETReg(ModelExtension):
    """
    Model architecture from https://arxiv.org/abs/2106.04550.
    Creates a :class:`~arcgis.learn.DETReg` object detection model,
    based on https://github.com/amirbar/DETReg.

    =============================   =============================================
    **Parameter**                    **Description**
    -----------------------------   ---------------------------------------------
    data                            Required fastai Databunch. Returned data object from
                                    :meth:`~arcgis.learn.prepare_data`  function.
    -----------------------------   ---------------------------------------------
    backbone                        Optional string. Backbone convolutional neural network
                                    model used for feature extraction, which
                                    is `resnet50` by default.
                                    Supported backbones: ResNet family.
    -----------------------------   ---------------------------------------------
    pretrained_path                 Optional string. Path where pre-trained model is
                                    saved.
    =============================   =============================================

    :return: :class:`~arcgis.learn.DETReg` Object
    """

    def __init__(self, data, backbone="resnet50", pretrained_path=None, **kwargs):
        self._check_dataset_support(data)
        backbone_name = backbone if type(backbone) is str else backbone.__name__
        if backbone_name not in self.supported_backbones:
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"
            )

        super().__init__(data, CustomDetReg, backbone, pretrained_path, **kwargs)

        idx = self._freeze()
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=3e-3)

    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def _freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            pass
        return idx

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "average_precision"]

    @property
    def supported_backbones(self):
        """Supported list of backbones for this model."""
        return DETReg._supported_backbones()

    @staticmethod
    def backbones():
        """Supported list of backbones for this model."""
        return DETReg._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return DETReg._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["PASCAL_VOC_rectangles", "KITTI_rectangles"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.DETReg` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.

        =====================   ===========================================

        :return: :class:`~arcgis.learn.DETReg` Object
        """
        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        backbone = emd["ModelParameters"]["backbone"]
        dataset_type = emd.get("DatasetType", "PASCAL_VOC_rectangles")
        chip_size = emd["ImageWidth"]
        resize_to = emd.get("resize_to", None)
        kwargs = emd.get("Kwargs", {})
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        data_passed = True
        if data is None:
            data_passed = False
            train_tfms = []
            val_tfms = []
            ds_tfms = (train_tfms, val_tfms)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                sd = ImageList([], path=emd_path.parent.parent.parent).split_by_idx([])
                data = (
                    sd.label_const(
                        0,
                        label_cls=ObjectDetectionCategoryList,
                        classes=list(class_mapping.values()),
                    )
                    .transform(ds_tfms)
                    .databunch(device=_get_device())
                    .normalize(imagenet_stats)
                )
            # Add 1 for background class
            data.c += 1
            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data.classes = ["background"] + list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = dataset_type

        data.resize_to = resize_to
        detreg = cls(data, backbone, pretrained_path=str(model_file), **kwargs)

        if not data_passed:
            detreg.learn.data.single_ds.classes = detreg._data.classes
            detreg.learn.data.single_ds.y.classes = detreg._data.classes

        return detreg

    def predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=False,
        visualize=False,
        resize=False,
    ):
        """
        Runs prediction on an Image. This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean. Will return the probability
                                scores of the bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the
                                same size (chip_size parameter in prepare_data)
                                that the model was trained on, before detecting
                                objects. Note that if resize_to parameter was
                                used in prepare_data, the image is resized to
                                that size instead.

                                By default, this parameter is false and the
                                detections are run in a sliding window fashion
                                by applying the model on cropped sections of
                                the image (of the same size as the model was
                                trained on).
        =====================   ===========================================

        :return: Returns a tuple with predictions, labels and optionally confidence scores
                  if return_scores=True. The predicted bounding boxes are returned as a list
                  of lists containing the  xmin, ymin, width and height of each predicted object
                  in each image. The labels are returned as a list of class values and the
                  confidence scores are returned as a list of floats indicating the confidence
                  of each prediction.
        """

    def predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            "assignment_iou_thrd": 0.3,
            "vanish_frames": 40,
            "detect_frames": 10,
        },
        visual_options={
            "show_scores": True,
            "show_labels": True,
            "thickness": 2,
            "fontface": 0,
            "color": (255, 255, 255),
        },
        resize=False,
    ):
        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.
        This method is only supported for RGB images.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path input_video_path
                                appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers,
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the video frames to the same size
                                (chip_size parameter in prepare_data) that the model was
                                trained on, before detecting objects. Note that if
                                resize_to parameter was used in prepare_data,
                                the video frames are resized to that size instead.

                                By default, this parameter is false and the detections
                                are run in a sliding window fashion by applying the
                                model on cropped sections of the frame (of the same
                                size as the model was trained on).
        =====================   ===========================================

        """

    def average_precision_score(
        self,
        detect_thresh=0.2,
        iou_thresh=0.1,
        mean=False,
        show_progress=True,
    ):
        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================

        :return: `dict` if mean is False otherwise `float`
        """

    def show_results(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================

        """



from __future__ import annotations
from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis import GIS
from typing import Dict, Any


class ExternalContentManager(_BaseKube):
    """
    Provides management of the external content resources.
    """

    _gis = None
    _con = None
    _properties = None
    _url = None

    def __init__(self, url: str, gis: GIS):
        super()
        if url.lower().endswith("/content") == False:
            url += "/content"
        self._url = url
        self._gis = gis
        self._con = gis._con

    @property
    def external_content(self) -> dict[str, Any]:
        """
        The external_content resource returns whether access to external
        content has been enabled or disabled for an organization. If the
        resource returns true, an organization's Esri content will contain
        external URLs that reference sites and resources hosted outside of
        the organization. If the resource returns false, Esri content
        containing external URLs will be removed from the organization. Any
        Esri content remaining after external content is disabled will not
        contain external URLs. If you are configuring ArcGIS Enterprise on
        Kubernetes in an environment in which no internet connection is
        available, or internet access is prohibited, access to external
        content should be disabled to avoid discovering content containing
        inaccessible URLs to external sites.

        :return: dict[str,Any]
        """
        url: str = f"{self._url}/externalcontent"
        params: dict[str, Any] = {
            "f": "json",
        }
        return self._con.get(url, params)

    def update(self, value: bool) -> dict[str, Any]:
        """
        The update operation enables and disables access to Esri-provided
        content containing external URLs that reference sites and resources
        hosted outside of the organization.

        :returns: dict[str,Any]
        """
        url: str = f"{self._url}/externalcontent/update"
        params: dict[str, Any] = {
            "f": "json",
            "externalContentEnabled": value,
        }
        return self._con.post(url, params)


###########################################################################
class LanguageManager(_BaseKube):
    """
    Provides management of the language resources. The languages resource
    provides a list of current languages for an organization.

    """

    _gis = None
    _con = None
    _properties = None
    _url = None

    def __init__(self, url: str, gis: GIS):
        super()
        if url.lower().endswith("/languages") == False:
            url += "/languages"
        self._url = url
        self._gis = gis
        self._con = gis._con

    @property
    def languages(self) -> Dict[str, bool]:
        """
        This resource returns a list of all Esri supported languages and
        their associated language codes, as well as the current status of
        the language, either enabled (true) or disabled (false). After your
        organization has been configured, English (language code en) will
        be the only enabled language by default. Additional languages can
        be enabled using the Add operation. Once enabled, the Esri provided
        content for these languages will be searchable and accessible to
        users in your organization. Enabled languages can be disabled using
        the Remove operation, which will remove their Esri provided content
        from the organization. Note that at least one language must always
        be enabled for your organization.

        :return: Dict[str, bool]
        """
        params = {"f": "json"}
        return self._con.get(self._url, params)

    @languages.setter
    def languages(self, languages: Dict[str, bool]):
        """
        This resource returns a list of all Esri supported languages and
        their associated language codes, as well as the current status of
        the language, either enabled (true) or disabled (false). After your
        organization has been configured, English (language code en) will
        be the only enabled language by default. Additional languages can
        be enabled using the Add operation. Once enabled, the Esri provided
        content for these languages will be searchable and accessible to
        users in your organization. Enabled languages can be disabled using
        the Remove operation, which will remove their Esri provided content
        from the organization. Note that at least one language must always
        be enabled for your organization.

        :return: Dict[str, bool]
        """
        existing_languages = self.languages
        adds = {}
        removes = {}
        for k in languages.keys():
            if (k in existing_languages and k in languages) and languages[
                k
            ] != existing_languages[k]:
                if languages[k]:
                    adds[k] = True
                else:
                    removes[k] = False
        if adds:
            url = f"{self._url}/add"
            params = {"f": "json", "languages": ",".join(list(adds.keys()))}
            res = self._con.post(url, params)
            del res
        if removes:
            url = f"{self._url}/remove"
            params = {"f": "json", "languages": ",".join(list(adds.keys()))}
            res = self._con.post(url, params)
            del res
        return self.languages



try:
    import os, sys, json
    import numpy as np
    import torch
    import torch.nn as nn
    import math
    from .util import scale_batch, variable_tile_size_check
    import cv2

    HAS_TORCH = True
except Exception as e:
    HAS_TORCH = False

import arcgis
from arcgis.learn import MaskRCNN
from arcgis.learn.models._maskrcnn_utils import predict_tta

try:
    import arcpy
except:
    pass


def calculate_rectangle_size_from_batch_size(batch_size):
    """
    calculate number of rows and cols to composite a rectangle given a batch size
    :param batch_size:
    :return: number of cols and number of rows
    """
    rectangle_height = int(math.sqrt(batch_size) + 0.5)
    rectangle_width = int(batch_size / rectangle_height)

    if rectangle_height * rectangle_width > batch_size:
        if rectangle_height >= rectangle_width:
            rectangle_height = rectangle_height - 1
        else:
            rectangle_width = rectangle_width - 1

    if (rectangle_height + 1) * rectangle_width <= batch_size:
        rectangle_height = rectangle_height + 1
    if (rectangle_width + 1) * rectangle_height <= batch_size:
        rectangle_width = rectangle_width + 1

    # swap col and row to make a horizontal rect
    if rectangle_height > rectangle_width:
        rectangle_height, rectangle_width = rectangle_width, rectangle_height

    if rectangle_height * rectangle_width != batch_size:
        return batch_size, 1

    return rectangle_height, rectangle_width


def convert_bounding_boxes_to_coord_list(bounding_boxes):
    """
    convert bounding box numpy array to python list of point arrays
    :param bounding_boxes: numpy array of shape [n, 4]
    :return: python array of point numpy arrays, each point array is in shape [4,2]
    """
    num_bounding_boxes = bounding_boxes.shape[0]
    bounding_box_coord_list = []
    for i in range(num_bounding_boxes):
        coord_array = np.empty(shape=(4, 2), dtype=float)
        coord_array[0][0] = bounding_boxes[i][0]
        coord_array[0][1] = bounding_boxes[i][1]

        coord_array[1][0] = bounding_boxes[i][0]
        coord_array[1][1] = bounding_boxes[i][3]

        coord_array[2][0] = bounding_boxes[i][2]
        coord_array[2][1] = bounding_boxes[i][3]

        coord_array[3][0] = bounding_boxes[i][2]
        coord_array[3][1] = bounding_boxes[i][1]

        bounding_box_coord_list.append([coord_array.tolist()])

    return bounding_box_coord_list


def get_tile_size(model_height, model_width, padding, batch_height, batch_width):
    """
    Calculate request tile size given model and batch dimensions
    :param model_height:
    :param model_width:
    :param padding:
    :param batch_width:
    :param batch_height:
    :return: tile height and tile width
    """
    tile_height = (model_height - 2 * padding) * batch_height
    tile_width = (model_width - 2 * padding) * batch_width

    return tile_height, tile_width


def tile_to_batch(
    pixel_block, model_height, model_width, padding, fixed_tile_size=True, **kwargs
):
    inner_width = model_width - 2 * padding
    inner_height = model_height - 2 * padding

    band_count, pb_height, pb_width = pixel_block.shape
    pixel_type = pixel_block.dtype

    if fixed_tile_size is True:
        batch_height = kwargs["batch_height"]
        batch_width = kwargs["batch_width"]
    else:
        batch_height = math.ceil((pb_height - 2 * padding) / inner_height)
        batch_width = math.ceil((pb_width - 2 * padding) / inner_width)

    batch = np.zeros(
        shape=(batch_width * batch_height, band_count, model_height, model_width),
        dtype=pixel_type,
    )
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        # pixel block might not be the shape (band_count, model_height, model_width)
        sub_pixel_block = pixel_block[
            :,
            y * inner_height : y * inner_height + model_height,
            x * inner_width : x * inner_width + model_width,
        ]
        sub_pixel_block_shape = sub_pixel_block.shape
        batch[
            b, :, : sub_pixel_block_shape[1], : sub_pixel_block_shape[2]
        ] = sub_pixel_block

    return batch, batch_height, batch_width


def batch_to_tile(batch, batch_height, batch_width):
    batch_size, bands, inner_height, inner_width = batch.shape
    tile = np.zeros(
        shape=(bands, inner_height * batch_height, inner_width * batch_width),
        dtype=batch.dtype,
    )

    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        tile[
            :,
            y * inner_height : (y + 1) * inner_height,
            x * inner_width : (x + 1) * inner_width,
        ] = batch[b]

    return tile


class ChildInstanceDetector:
    def initialize(self, model, model_as_file):
        if not HAS_TORCH:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/"
            )

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info["ModelFile"]
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        self.model_emd = model

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "padding",
                    "dataType": "numeric",
                    "value": int(self.json_info["ImageHeight"]) // 4,
                    "required": False,
                    "displayName": "Padding",
                    "description": "Padding",
                },
                {
                    "name": "batch_size",
                    "dataType": "numeric",
                    "required": False,
                    "value": 4,
                    "displayName": "Batch Size",
                    "description": "Batch Size",
                },
                {
                    "name": "threshold",
                    "dataType": "numeric",
                    "required": False,
                    "value": 0.9,
                    "displayName": "Threshold",
                    "description": "Threshold",
                },
                {
                    "name": "return_bboxes",
                    "dataType": "string",
                    "required": False,
                    "domain": ("True", "False"),
                    "value": "False",
                    "displayName": "return_bboxes",
                    "description": "return_bboxes",
                },
                {
                    "name": "test_time_augmentation",
                    "dataType": "string",
                    "required": False,
                    "value": "False",
                    "displayName": "Perform test time augmentation while predicting",
                    "description": "If True, will merge predictions from flipped and rotated images.",
                },
                {
                    "name": "merge_policy",
                    "dataType": "string",
                    "required": False,
                    "value": "mean",
                    "displayName": "Policy for merging augmented predictions",
                    "description": "Policy for merging predictions('mean' or 'nms'). Applicable when test_time_augmentation is True.",
                },
            ]
        )
        required_parameters = variable_tile_size_check(
            self.json_info, required_parameters
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.tytx = int(scalars.get("tile_size", self.json_info["ImageHeight"]))
        self.mask_rcnn = MaskRCNN.from_model(
            emd_path=self.model_emd, chip_size=self.tytx
        )
        self._learnmodel = self.mask_rcnn
        self.model = self.mask_rcnn.learn.model.to(self.device)
        self.model.eval()

        #
        self.padding = int(
            scalars.get("padding", self.tytx // 4)
        )  ## Default padding Imageheight//4.
        self.batch_size = (
            int(math.sqrt(int(scalars.get("batch_size", 4)))) ** 2
        )  ## Default 4 batch_size
        self.threshold = float(scalars.get("threshold", 0.9))  ## Default 0.9 threshold.
        self.return_bboxes = eval(scalars.get("return_bboxes", "False"))
        self.merge_policy = scalars.get("merge_policy", "mean").lower()
        self.use_tta = scalars.get("test_time_augmentation", "false").lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]

        if self.use_tta:
            if self.json_info["ImageSpaceUsed"] == "MAP_SPACE":
                self.model.arcgis_tta = list(range(8))
            else:
                self.model.arcgis_tta = [0, 2]

        (
            self.rectangle_height,
            self.rectangle_width,
        ) = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(
            self.tytx,
            self.tytx,
            self.padding,
            self.rectangle_height,
            self.rectangle_width,
        )

        return {
            "extractBands": tuple(self.json_info["ExtractBands"]),
            "padding": self.padding,
            "tx": tx,
            "ty": ty,
            "fixedTileSize": 1,
        }

    def vectorize(self, **pixelBlocks):  # 8 x 3 x 224 x 224
        input_image = pixelBlocks["raster_pixels"].astype(np.float32)
        batch, batch_height, batch_width = tile_to_batch(
            input_image,
            self.tytx,
            self.tytx,
            self.padding,
            fixed_tile_size=True,
            batch_height=self.rectangle_height,
            batch_width=self.rectangle_width,
        )

        if "NormalizationStats" in self.json_info:
            img_normed = scale_batch(batch, self.json_info)
        else:
            img_normed = batch / 255

        predictions = pixel_mask_image(
            self.model,
            img_normed,
            self.device,
            self.tytx,
            threshold=self.threshold,
            batch_size=self.batch_size,
            return_bboxes=self.return_bboxes,
            use_tta=self.use_tta,
            merge_policy=self.merge_policy,
        )

        return predictions


def predict_mask_rcnn(
    model, images, device, chip_size, threshold=0.5, use_tta=False, merge_policy="mean"
):
    model = model.to(device)
    normed_batch_tensor = torch.tensor(images).to(device).float()
    if use_tta:
        predictions = predict_tta(model, normed_batch_tensor, threshold, merge_policy)
    else:
        temp = model.roi_heads.score_thresh
        model.roi_heads.score_thresh = threshold
        predictions = model(list(normed_batch_tensor))
        model.roi_heads.score_thresh = temp

    return predictions


def pixel_mask_image(
    model,
    img_normed,
    device,
    chip_size,
    threshold=0.5,
    batch_size=4,
    return_bboxes=False,
    use_tta=False,
    merge_policy="mean",
):
    side = int(math.sqrt(batch_size))

    predictions = predict_mask_rcnn(
        model,
        img_normed,
        device,
        chip_size,
        threshold=threshold,
        use_tta=use_tta,
        merge_policy=merge_policy,
    )

    all_contour_list = []
    pred_box = []
    pred_class = []
    pred_score = []

    for batch_idx in range(len(predictions)):
        i, j = batch_idx // side, batch_idx % side
        masks = predictions[batch_idx]["masks"].squeeze().detach().cpu().numpy()
        if masks.shape[0] != 0:  # handle for prediction with n masks
            if (
                len(masks.shape) == 2
            ):  # for mask dimension hxw (in case of only one predicted mask)
                masks = masks[None]
            for n, mask in enumerate(masks):
                if predictions[batch_idx]["scores"][n].tolist() >= threshold:
                    contours, hierarchy = cv2.findContours(
                        (mask >= 0.5).astype(np.uint8),
                        cv2.RETR_TREE,
                        cv2.CHAIN_APPROX_NONE,
                    )
                    if len(contours) > 0:
                        hierarchy = hierarchy[0]
                        for c_idx, contour in enumerate(contours):
                            contour = contours[c_idx] = contour.squeeze(1)
                            contours[c_idx][:, 0] = contour[:, 0] + (j * chip_size)
                            contours[c_idx][:, 1] = contour[:, 1] + (i * chip_size)
                        for (
                            contour_idx,
                            (next_contour, prev_contour, child_contour, parent_contour),
                        ) in enumerate(hierarchy):
                            if parent_contour == -1:
                                coord_list = []
                                # check if it is a state line
                                closed_contour = (
                                    all(
                                        contours[contour_idx].max(axis=0)
                                        - contours[contour_idx].min(axis=0)
                                    )
                                    and contours[contour_idx].shape[0] > 2
                                )
                                if closed_contour:
                                    coord_list.append(contours[contour_idx].tolist())
                                while child_contour != -1:
                                    closed_contour = (
                                        all(
                                            contours[child_contour].max(axis=0)
                                            - contours[child_contour].min(axis=0)
                                        )
                                        and contours[child_contour].shape[0] > 2
                                    )
                                    if closed_contour:
                                        coord_list.append(
                                            contours[child_contour].tolist()
                                        )
                                    child_contour = hierarchy[child_contour][0]
                                #
                                if coord_list != []:
                                    all_contour_list.append(coord_list)
                                    pred_class.append(
                                        predictions[batch_idx]["labels"][n].tolist()
                                    )
                                    pred_score.append(
                                        predictions[batch_idx]["scores"][n].tolist()
                                        * 100
                                    )
                                    box = (
                                        predictions[batch_idx]["boxes"][n]
                                        .cpu()
                                        .detach()
                                        .numpy()
                                    )
                                    box[0] += j * chip_size
                                    box[2] += j * chip_size
                                    box[1] += i * chip_size
                                    box[3] += i * chip_size
                                    pred_box.append(box)

    if return_bboxes:
        pred_box = np.array(pred_box)
        bbox = convert_bounding_boxes_to_coord_list(pred_box)
        return bbox, pred_class, pred_score

    else:
        return all_contour_list, pred_class, pred_score
